"""
Downloading input files from URIs, with plugin modules for different URI schemes

Download URI plugins are installed & registered using the setuptools entry point group
"miniwdl.plugin.file_download", with name equal to the URI scheme (e.g. "gs" or "s3").

The plugin entry point should be a context manager, which the runtime keeps open for the duration of
the download operation. Given the desired URI, it should quickly yield a tuple with:
    1. source code of a WDL 1.0 task to perform the download
    2. dict of Cromwell-style JSON inputs to give to the task
miniwdl then executes this specified operation, expecting it to produce an output "File file" with
the downloaded file. By doing the heavy lifting in a WDL task, the operation gets to inherit all
the functionality of miniwdl's task runtime, e.g. pulling docker image with binary dependencies,
resource scheduling & isolation, logging, error/signal handling, retry, etc.
The Python context manager itself might be used to obtain and manage the lifetime of any needed
security credentials.
"""
import os
import logging
import traceback
import tempfile
import hashlib
from contextlib import ExitStack
from typing import Optional, Generator, Dict, Any, Tuple, Callable
from . import config
from .cache import CallCache
from .._util import compose_coroutines
from .._util import StructuredLogMessage as _


def _load(cfg: config.Loader):
    if getattr(cfg, "_downloaders", None):
        return

    # default public URI downloaders
    file_downloaders = {
        "https": aria2c_downloader,
        "http": aria2c_downloader,
        "ftp": aria2c_downloader,
    }
    directory_downloaders = {}

    # plugins
    for plugin_name, plugin_fn in config.load_plugins(cfg, "file_download"):
        file_downloaders[plugin_name] = plugin_fn
    for plugin_name, plugin_fn in config.load_plugins(cfg, "directory_download"):
        directory_downloaders[plugin_name] = plugin_fn

    setattr(cfg, "_downloaders", (file_downloaders, directory_downloaders))


def _downloader(
    cfg: config.Loader, uri: str, directory: bool = False
) -> Optional[Callable[..., Generator[Dict[str, Any], Dict[str, Any], None]]]:
    _load(cfg)
    colon = uri.find(":")
    if colon <= 0:
        return None
    scheme = uri[:colon]
    return getattr(cfg, "_downloaders")[1 if directory else 0].get(scheme, None)


def able(cfg: config.Loader, uri: Optional[str], directory: bool = False) -> bool:
    """
    Returns True if uri appears to be a URI we know how to download
    """
    return bool(uri and _downloader(cfg, uri, directory=directory) is not None)


def run(
    cfg: config.Loader, logger: logging.Logger, uri: str, directory: bool = False, **kwargs
) -> str:
    """
    Download the URI and return the local filename.

    kwargs are passed through to ``run_local_task``, so ``run_dir`` and ``logger_prefix`` may be
    useful in particular.
    """

    from .error import RunFailed, DownloadFailed, Terminated, error_json
    from .task import run_local_task
    from .. import parse_document, values_from_json, values_to_json, Walker

    gen = _downloader(cfg, uri, directory=directory)
    assert gen
    try:
        logger.info(_(f"start {'directory ' if directory else ''}download", uri=uri))
        with compose_coroutines([lambda kwargs: gen(cfg, logger, **kwargs)], {"uri": uri}) as cor:
            recv = next(cor)

            if "task_wdl" in recv:
                task_wdl, inputs = (recv[k] for k in ["task_wdl", "inputs"])

                doc = parse_document(task_wdl, version="development")  # pyre-ignore
                assert len(doc.tasks) == 1 and not doc.workflow
                doc.typecheck()
                Walker.SetParents()(doc)
                task = doc.tasks[0]
                inputs = values_from_json(inputs, task.available_inputs)  # pyre-ignore
                subdir, outputs_env = run_local_task(
                    cfg, task, inputs, run_id=("download-" + task.name), **kwargs
                )

                recv = cor.send(
                    {"outputs": values_to_json(outputs_env), "dir": subdir}  # pyre-ignore
                )

            ans = recv["outputs"]["directory" if directory else "file"]
            assert isinstance(ans, str) and os.path.exists(ans)
            logger.notice(  # pyre-ignore
                _(f"downloaded{' directory' if directory else ' file'}", uri=uri, file=ans)
            )
            return ans

    except RunFailed as exn:
        if isinstance(exn.__cause__, Terminated):
            raise exn.__cause__ from None
        raise DownloadFailed(uri) from exn.__cause__
    except Exception as exn:
        logger.debug(traceback.format_exc())
        logger.error(_("downloader error", uri=uri, **error_json(exn)))
        raise DownloadFailed(uri) from exn


def run_cached(
    cfg,
    logger: logging.Logger,
    cache: CallCache,
    uri: str,
    run_dir: str,
    directory: bool = False,
    **kwargs,
) -> Tuple[bool, str]:
    """
    Cached download logic: returns the file from the cache if available; otherwise, runs the
    download and puts it into the cache before returning
    """
    cached = cache.get_download(uri, directory=directory, logger=logger)
    if cached:
        return True, cached
    cache_path = cache.download_cacheable(uri, directory=directory)
    cache_path_preexists = cache_path and os.path.exists(cache_path)
    if cache_path and not cache_path_preexists:
        # run the download in a holding area under the cache directory, so that it can later be
        # moved atomically into its final cache location
        run_dir = os.path.join(cfg["download_cache"]["dir"], "ops")
    filename = run(cfg, logger, uri, directory=directory, run_dir=run_dir, **kwargs)
    if cache_path_preexists:
        # a cache entry had already existed, but we didn't use it (--no-cache).
        # FIXME: it'd be better to replace the old copy...but what if another workflow is using it?
        logger.warning(
            _(
                "ignored a previously-cached download, which remains in the cache",
                uri=uri,
                downloaded=filename,
                cache_path=cache_path,
            )
        )
        # use the newly downloaded copy in the current run directory
        cache.memo_download(uri, filename, directory=directory)
        return False, filename
    return False, cache.put_download(
        uri, os.path.realpath(filename), directory=directory, logger=logger
    )


# WDL tasks for downloading a file based on its URI scheme


def aria2c_downloader(
    cfg: config.Loader, logger: logging.Logger, uri: str, **kwargs
) -> Generator[Dict[str, Any], Dict[str, Any], None]:
    wdl = r"""
    task aria2c {
        input {
            String uri
            String docker
            Int connections = 10
        }
        command <<<
            set -euxo pipefail
            mkdir __out
            cd __out
            aria2c -x ~{connections} -s ~{connections} \
                --file-allocation=none --retry-wait=2 --stderr=true --enable-color=false \
                "~{uri}"
        >>>
        output {
            File file = glob("__out/*")[0]
        }
        runtime {
            cpu: 4
            memory: "1G"
            docker: docker
        }
    }
    """
    recv = yield {
        "task_wdl": wdl,
        "inputs": {"uri": uri, "docker": cfg["download_aria2c"]["docker"]},
    }
    yield recv  # pyre-ignore


def awscli_downloader(
    cfg: config.Loader, logger: logging.Logger, uri: str, **kwargs
) -> Generator[Dict[str, Any], Dict[str, Any], None]:
    inputs: Dict[str, Any] = {"uri": uri, "docker": cfg["download_awscli"]["docker"]}
    with ExitStack() as cleanup:
        inputs["aws_credentials"] = prepare_aws_credentials(cfg, logger, cleanup)

        wdl = r"""
        task aws_s3_cp {
            input {
                String uri
                String docker
                File? aws_credentials
            }

            command <<<
                set -euo pipefail
                if [ -n "~{aws_credentials}" ]; then
                    source "~{aws_credentials}"
                fi
                set -x
                mkdir __out
                if ! aws s3 cp "~{uri}" __out/ ; then
                    # Retry with --no-sign-request in case the object is public. Without this flag,
                    # the previous invocation could have failed either because (i) no AWS
                    # credentials are available or (ii) the IAM policy restricts accessible S3
                    # buckets regardless of whether the desired object is public.
                    rm -f __out/*
                    aws s3 cp --no-sign-request "~{uri}" __out/
                fi
            >>>

            output {
                File file = glob("__out/*")[0]
            }

            runtime {
                cpu: 4
                memory: "1G"
                docker: docker
            }
        }
        """
        recv = yield {"task_wdl": wdl, "inputs": inputs}
    yield recv  # pyre-ignore


def awscli_directory_downloader(
    cfg: config.Loader, logger: logging.Logger, uri: str, **kwargs
) -> Generator[Dict[str, Any], Dict[str, Any], None]:
    inputs: Dict[str, Any] = {"uri": uri, "docker": cfg["download_awscli"]["docker"]}
    with ExitStack() as cleanup:
        inputs["aws_credentials"] = prepare_aws_credentials(cfg, logger, cleanup)

        wdl = r"""
        task aws_s3_cp_directory {
            input {
                String uri
                String docker
                File? aws_credentials
            }

            String dnm = basename(uri, "/")

            command <<<
                set -euo pipefail
                if [ -n "~{aws_credentials}" ]; then
                    source "~{aws_credentials}"
                fi
                set -x
                mkdir -p "__out/~{dnm}/"
                if ! aws s3 cp --recursive "~{uri}" "__out/~{dnm}/" ; then
                    # Retry with --no-sign-request in case the object is public. Without this flag,
                    # the previous invocation could have failed either because (i) no AWS
                    # credentials are available or (ii) the IAM policy restricts accessible S3
                    # buckets regardless of whether the desired object is public.
                    rm -f "__out/~{dnm}/*"
                    aws s3 cp --recursive --no-sign-request "~{uri}" "__out/~{dnm}/"
                fi
            >>>

            output {
                Directory directory = "__out/" + dnm
            }

            runtime {
                cpu: 4
                memory: "1G"
                docker: docker
            }
        }
        """
        recv = yield {"task_wdl": wdl, "inputs": inputs}
    yield recv  # pyre-ignore


def prepare_aws_credentials(
    cfg: config.Loader, logger: logging.Logger, cleanup: ExitStack
) -> Optional[str]:
    host_aws_credentials = {}
    if "AWS_EC2_METADATA_DISABLED" in os.environ:
        # https://github.com/aws/aws-cli/issues/5623
        host_aws_credentials["AWS_EC2_METADATA_DISABLED"] = os.environ["AWS_EC2_METADATA_DISABLED"]
    # get AWS credentials from boto3 (unless prevented by configuration)
    if cfg["download_awscli"].get_bool("host_credentials"):
        try:
            import boto3  # pyre-fixme

            b3creds = boto3.session.Session().get_credentials()
            host_aws_credentials["AWS_ACCESS_KEY_ID"] = b3creds.access_key
            host_aws_credentials["AWS_SECRET_ACCESS_KEY"] = b3creds.secret_key
            host_aws_credentials["AWS_SESSION_TOKEN"] = b3creds.token
        except Exception:
            pass

    if host_aws_credentials:
        # write credentials to temp file that'll self-destruct afterwards
        host_aws_credentials = (
            "\n".join(f"export {k}='{v}'" for (k, v) in host_aws_credentials.items()) + "\n"
        )
        aws_credentials_file = cleanup.enter_context(
            tempfile.NamedTemporaryFile(
                prefix=hashlib.sha256(host_aws_credentials.encode()).hexdigest(),
                delete=True,
                mode="w",
            )
        )
        print(host_aws_credentials, file=aws_credentials_file, flush=True)
        # make file group-readable to ensure it'll be usable if the docker image runs as non-root
        os.chmod(aws_credentials_file.name, os.stat(aws_credentials_file.name).st_mode | 0o40)
        logger.getChild("awscli_downloader").info("loaded host AWS credentials")
        return aws_credentials_file.name
    else:
        logger.getChild("awscli_downloader").info(
            "no AWS credentials available via host awscli/boto3; if needed, "
            "configure them and set [download_awscli] host_credentials=true. "
            "(On EC2: awscli might still assume role from instance metadata "
            "service.)"
        )
        return None


def gsutil_downloader(
    cfg: config.Loader, logger: logging.Logger, uri: str, **kwargs
) -> Generator[Dict[str, Any], Dict[str, Any], None]:
    """
    Built-in downloader plugin for public gs:// URIs; registered by setup.cfg entry_points section

    TODO: adopt security credentials from runtime environment
    """
    if uri == "gs://8675309":
        # hook for test coverage of exception handler
        raise RuntimeError("don't change your number")
    wdl = r"""
    task gsutil_cp {
        input {
            String uri
            String docker
        }
        command <<<
            set -euxo pipefail
            mkdir __out/
            gsutil -q cp "~{uri}" __out/
        >>>
        output {
            File file = glob("__out/*")[0]
        }
        runtime {
            cpu: 4
            memory: "1G"
            docker: docker
        }
    }
    """
    yield (  # pyre-ignore
        yield {"task_wdl": wdl, "inputs": {"uri": uri, "docker": cfg["download_gsutil"]["docker"]}}
    )



# pyre-strict
from typing import Union, Dict, Any, Optional
from ..Error import RuntimeError as _RuntimeError, SourcePosition
from ..Tree import Task as _Task, Workflow as _Workflow
from . import _statusbar


class CommandFailed(_RuntimeError):
    """
    Failure of the task command
    """

    exit_status: int
    """
    Task command exit status
    """

    stderr_file: str
    """
    Path to a file containing the task's standard error
    """

    stdout_file: str
    """
    Path to a file containing the task's standard output
    """

    def __init__(
        self,
        exit_status: int,
        stderr_file: str,
        stdout_file: str,
        message: str = "",
        **kwargs,  # pyre-ignore
    ) -> None:
        oom_hint = ", a possible indication that it ran out of memory" if exit_status == 137 else ""
        super().__init__(
            message or f"task command failed with exit status {exit_status}{oom_hint}", **kwargs
        )
        self.exit_status = exit_status
        self.stderr_file = stderr_file
        self.stdout_file = stdout_file


class Terminated(_RuntimeError):
    """
    Workflow/task was intentionally terminated, e.g. by Unix signal
    """

    quiet: bool
    """
    Termination warrants less logging because it was a secondary side-effect of a previous error
    """

    def __init__(self, quiet: bool = False, **kwargs) -> None:  # pyre-ignore
        super().__init__(**kwargs)
        self.quiet = quiet
        _statusbar.abort()


class Interrupted(_RuntimeError):
    """
    Task was interrupted by an exogenous problem (e.g. worker node went down)
    """

    pass


class OutputError(_RuntimeError):
    """
    Failure whilst gathering task outputs
    """

    pass


class DownloadFailed(_RuntimeError):
    """
    Failure to download a URI input file
    """

    uri: str

    def __init__(self, uri: str, message: str = "") -> None:
        super().__init__(message or ("unable to download " + uri))
        self.uri = uri
        _statusbar.abort()


class RunFailed(_RuntimeError):
    """"""

    exe: Union[_Task, _Workflow]
    run_id: str
    run_dir: str

    # pyre-ignore
    def __init__(self, exe: Union[_Task, _Workflow], run_id: str, run_dir: str, **kwargs) -> None:
        super().__init__(
            f"{'task' if isinstance(exe, _Task) else 'workflow'} {exe.name} "
            f"({exe.pos.uri} Ln {exe.pos.line} Col {exe.pos.column}) failed",
            **kwargs,
        )
        self.exe = exe
        self.run_id = run_id
        self.run_dir = run_dir
        _statusbar.abort()


def error_json(exn: BaseException, cause: Optional[Exception] = None) -> Dict[str, Any]:
    """
    Make a json-dumpable dict to write into error sentinel file
    """

    def pos_json(pos: SourcePosition) -> Dict[str, Any]:
        return {"source": pos.abspath, "line": pos.line, "column": pos.column}

    info: Dict[str, Any] = {"error": exn.__class__.__name__}
    pos = None
    if isinstance(exn, RunFailed):
        exe = getattr(exn, "exe")
        pos = getattr(exe, "pos")
        info["task" if isinstance(exe, _Task) else "workflow"] = exe.name
        info["run"] = getattr(exn, "run_id")
        info["dir"] = getattr(exn, "run_dir")
        # follow __cause__s to find the original triggering exception
        from_exn = exn
        from_run = None
        from_dir = dir
        from_pos = pos
        while isinstance(from_exn, RunFailed):
            from_pos = getattr(getattr(from_exn, "exe"), "pos", from_pos)
            from_dir = getattr(from_exn, "run_dir")
            from_run = getattr(from_exn, "run_id")
            from_exn = cause or from_exn.__cause__
            from_pos = getattr(from_exn, "pos", from_pos)
            cause = None
        if from_exn and from_exn is not exn:
            info["cause"] = error_json(from_exn)
            info["cause"]["run"] = from_run
            info["cause"]["dir"] = from_dir
            info["cause"]["pos"] = pos_json(from_pos)
    elif isinstance(exn, CommandFailed):
        info["exit_status"] = getattr(exn, "exit_status")
        info["stderr_file"] = getattr(exn, "stderr_file")
        info["stdout_file"] = getattr(exn, "stdout_file")
    elif str(exn):
        info["message"] = str(exn)
    if hasattr(exn, "job_id"):
        info["node"] = getattr(exn, "job_id")
    pos = pos or getattr(exn, "pos", None)
    if isinstance(pos, SourcePosition):
        info["pos"] = pos_json(pos)
    if isinstance(getattr(exn, "more_info", None), dict):
        more_info = getattr(exn, "more_info")
        for k in more_info:
            if k not in info:
                info[k] = more_info[k]
    return info



"""
`miniwdl <https://github.com/chanzuckerberg/miniwdl/>`_ is a developer toolkit and local runner for
the bioinformatics-focused `Workflow Description Language (WDL) <http://openwdl.org/>`_. This
documentation covers the Python3 ``WDL`` package facilitating parsing & static analysis of WDL
documents. Simply ``import WDL`` once miniwdl has been installed.

* `GitHub repo <https://github.com/chanzuckerberg/miniwdl/>`_ for installation and further background
* `Codelabs <https://miniwdl.readthedocs.io/en/latest/WDL.html#python-codelabs>`_ on using this package
"""
import sys
import os
import errno
import inspect
from typing import List, Optional, Callable, Dict, Any, Awaitable, Union
from . import _util, _parser, Error, Type, Value, Env, Expr, Tree, Walker
from .Tree import (
    Decl,
    StructTypeDef,
    Task,
    Call,
    Scatter,
    Conditional,
    Gather,
    Workflow,
    Document,
    WorkflowNode,
    WorkflowSection,
    SourceComment,
)

SourcePosition = Error.SourcePosition
SourceNode = Error.SourceNode


def load(
    uri: str,
    path: Optional[List[str]] = None,
    check_quant: bool = True,
    read_source: Optional[
        Callable[[str, List[str], Optional[Document]], Awaitable["ReadSourceResult"]]
    ] = None,
    import_max_depth: int = 10,
) -> Document:
    """
    Parse a WDL document given filename/URI, recursively descend into imported documents, then typecheck the tasks and
    workflow.

    :param path: local filesystem directories to search for imports, in addition to the current working directory

    :param check_quant:
        set to ``False`` to relax static typechecking of the optional (?) and nonempty (+) type quantifiers. This is
        discouraged, but may be useful for older WDL workflows which assume less-rigorous static validation of these
        annotations.

    :param read_source:
        async routine to read the WDL source code from filename/URI; see :func:`read_source_default` below for details

    :param import_max_depth:
        to prevent recursive import infinite loops, fail when there are too many import nesting levels (default 10)

    :raises WDL.Error.SyntaxError: when the document is syntactically invalid under the WDL grammar
    :raises WDL.Error.ValidationError:
        when the document is syntactically OK, but fails typechecking or other static validity checks
    :raises WDL.Error.MultipleValidationErrors:
        when multiple validation errors are detected in one pass, listed in the ``exceptions`` attribute
    :raises WDL.Error.ImportError:
        when an imported sub-document can't be loaded; the ``__cause__`` attribute has the specific error
    """
    doc = Tree._load(
        uri,
        path=path,
        check_quant=check_quant,
        read_source=read_source,
        import_max_depth=import_max_depth,
    )
    Walker.SetParents()(doc)
    return doc


async def load_async(
    uri: str,
    path: Optional[List[str]] = None,
    check_quant: bool = True,
    read_source: Optional[
        Callable[[str, List[str], Optional[Document]], Awaitable["ReadSourceResult"]]
    ] = None,
    import_max_depth: int = 10,
) -> Document:
    """
    Async version of :func:`load`, with all the same arguments
    """
    doc = await Tree._load_async(
        uri,
        path=path,
        check_quant=check_quant,
        read_source=read_source,
        import_max_depth=import_max_depth,
    )
    Walker.SetParents()(doc)
    return doc


async def read_source_default(
    uri: str, path: List[str], importer: Optional[Document]
) -> "ReadSourceResult":
    """
    Default async routine for the ``read_source`` parameter to :func:`load` and :func:`load_async`,
    which they use to read the desired WDL document and its imports. This default routine handles
    local files only, supplying the search path logic to resolve relative filenames; it fails with
    network URIs.

    :param uri: Filename/URI to read, as provided to :func:`load` or the WDL import statement; may
                be relative
    :param path: Local directories to search for relative imports
    :param importer: The document importing the one here requested, if any; the
                     ``importer.pos.uri`` and ``importer.pos.abspath`` fields may be relevant to
                     resolve relative imports.
    :returns: ``ReadSourceResult(source_text="...", abspath="...")``

    Callers may wish to override ``read_source`` with logic to download source code from network
    URIs, and for local filenames fall back to ``return await WDL.read_source_default(...)``.

    Note: the synchronous :func:`load` merely calls :func:`load_async` on the current
    ``asyncio.get_event_loop()`` and awaits the result.
    """
    return await Tree.read_source_default(uri, path, importer)


class ReadSourceResult(Tree.ReadSourceResult):
    """
    The ``NamedTuple`` to be returned by the ``read_source`` routine. Its ``source_text: str`` field
    provides the WDL source code, and the ``abspath: str`` field is the absolute filename/URI from
    which the source was read (e.g. after resolving a relative path).
    """


async def resolve_file_import(uri: str, path: List[str], importer: Optional[Document]) -> str:
    """
    Exposes the logic by which :func:`read_source_default` resolves ``uri`` to the absolute path of
    an extant file. If ``uri`` is already an absolute path, it's normalized and returned. A
    relative ``uri`` is resolved by first joining it to either, the directory of the importer
    document (if any), or the process current working directory (otherwise). Failing that, it's
    searched in the ``path`` directories (in reverse order).

    Security-focused applications may wish to override ``read_source`` with logic to restrict
    allowable results of ``resolve_file_import``, to prevent WDL source code from triggering access
    to arbitrary filesystem paths. No such restrictions are applied by default.
    """
    return await Tree.resolve_file_import(uri, path, importer)


def copy_source(doc: Document, dir: str) -> str:
    ""
    """
    Copy the original WDL document source, and any imports, into the specified directory. Ignores
    any imports using absolute file paths or URIs.

    Returns the path to the copy of the given document, which possibly could be nested in a
    subdirectory if it uses .. relative imports.
    """
    # make list of all docs to save
    docs = []
    queue = [doc]
    while queue:
        a_doc = queue.pop()
        docs.append(a_doc)
        for imp in a_doc.imports:
            if (
                not imp.uri.startswith("http:")
                and not imp.uri.startswith("https:")
                and not os.path.isabs(imp.uri)
            ):
                queue.append(imp.doc)
    # find longest common prefix (up to a '/') among docs' pos.abspath (note these could be URIs!)
    lcp = os.path.dirname(os.path.commonprefix([a_doc.pos.abspath for a_doc in docs]))
    # write each doc text out under dir, its abspath without lcp
    ans = None
    for a_doc in docs:
        assert a_doc.pos.abspath.startswith(lcp)
        rp = a_doc.pos.abspath[len(lcp) :].lstrip("/")
        fn = os.path.join(dir, rp)
        os.makedirs(os.path.dirname(fn), exist_ok=True)
        _util.write_atomic(a_doc.source_text, fn, end="")
        if a_doc is doc:
            assert not ans
            ans = fn
    assert ans
    return ans


def parse_document(txt: str, version: Optional[str] = None, uri: str = "") -> Document:
    ""
    """
    Parse WDL document text into an abstract syntax tree. Doesn't descend into
    imported documents nor typecheck the AST.

    :param version:
        Override the WDL language version, such as "1.0" or "draft-2". (By default, detects from the "version" string
        at the beginning of the document, per the WDL spec.)

    :param uri: filename/URI for error reporting (not otherwise used)
    """
    doc = _parser.parse_document(txt, version, uri)
    Walker.SetParents()(doc)
    return doc


def parse_expr(txt: str, version: Optional[str] = None) -> Expr.Base:
    ""
    """
    Parse an isolated WDL expression text into an abstract syntax tree
    """
    return _parser.parse_expr(txt, version)


def parse_tasks(txt: str, version: Optional[str] = None) -> List[Task]:
    return _parser.parse_tasks(txt, version)


def values_from_json(
    values_json: Dict[str, Any],
    available: Env.Bindings[Union[Tree.Decl, Type.Base]],
    required: Optional[Env.Bindings[Union[Tree.Decl, Type.Base]]] = None,
    namespace: str = "",
) -> Env.Bindings[Value.Base]:
    """
    Given a dict parsed from Cromwell-style JSON and the available input (or
    output) declarations of a task or workflow, create a
    ``WDL.Env.Bindings[Value.Base]``.

    :param required: raise an error if any of these required inputs aren't
                     present
    :param namespace: expect each key to start with this namespace prefixed to
                      the input/output names (e.g. the workflow name)
    """
    if namespace and not namespace.endswith("."):
        namespace += "."
    ans = Env.Bindings()
    for key in values_json:
        if not key.startswith("#"):  # ignore "comments"
            key2 = key
            if namespace and key.startswith(namespace):
                key2 = key[len(namespace) :]

            ty = None
            if key2 in available:
                ty = available[key2]
            else:
                key2parts = key2.split(".")

                runtime_idx = next(
                    (i for i, term in enumerate(key2parts) if term in ("runtime",)), -1
                )
                if (
                    runtime_idx >= 0
                    and len(key2parts) > (runtime_idx + 1)
                    and ".".join(key2parts[:runtime_idx] + ["_runtime"]) in available
                ):
                    # allow arbitrary keys for runtime
                    ty = Type.Any()
                elif len(key2parts) == 3 and key2parts[0] and key2parts[1] and key2parts[2]:
                    # attempt to simplify <call>.<subworkflow>.<input> from old Cromwell JSON
                    key2 = ".".join([key2parts[0], key2parts[2]])
                    if key2 in available:
                        ty = available[key2]
            if not ty:
                raise Error.InputError("unknown input/output: " + key) from None
            if isinstance(ty, Tree.Decl):
                # treat input with default as optional, with or without the ? type quantifier
                ty = ty.type.copy(optional=True) if ty.expr else ty.type

            assert isinstance(ty, Type.Base)
            try:
                ans = ans.bind(key2, Value.from_json(ty, values_json[key]))
            except Error.InputError as exn:
                raise Error.InputError(exn.args[0] + f" (in {key})").with_traceback(
                    sys.exc_info()[2]
                )
    if required:
        missing = required.subtract(ans)
        if missing:
            raise Error.InputError(
                "missing required inputs/outputs: " + ", ".join(values_to_json(missing))
            )
    return ans


def values_to_json(values_env: Env.Bindings[Value.Base], namespace: str = "") -> Dict[str, Any]:
    """
    Convert a ``WDL.Env.Bindings[WDL.Value.Base]`` to a dict which ``json.dumps`` to
    Cromwell-style JSON.

    :param namespace: prefix this namespace to each key (e.g. workflow name)
    """
    # also can be used on Env.Bindings[Tree.Decl] or Env.Types, then the right-hand side of
    # each entry will be the type string.
    if namespace and not namespace.endswith("."):
        namespace += "."
    ans = {}
    for item in values_env:
        v = item.value
        if isinstance(v, Value.Base):
            j = v.json
        elif isinstance(item.value, Tree.Decl):
            j = str(item.value.type)
        else:
            assert isinstance(item.value, Type.Base)
            j = str(item.value)
        ans[(namespace if not item.name.startswith("_") else "") + item.name] = j
    return ans



# Portions copied from AirflowConfigParser --
#  https://github.com/apache/airflow/blob/master/airflow/configuration.py
# Exposition --
#  https://medium.com/@tszumowski/delightful-designs-airflows-configuration-parser-1ef1a6b3d03c


import os
import configparser
import logging
import json
from fnmatch import fnmatchcase
from typing import Optional, List, Dict, Any, Callable, TypeVar, Set, Tuple, Iterable
from xdg import XDG_CONFIG_DIRS, XDG_CONFIG_HOME
from .._util import StructuredLogMessage as _

_T = TypeVar("_T")


class ConfigMissing(Exception):
    # configparser.No{Option,Section}Error have vague hard-coded error messages
    pass


class Section:
    _section: str
    _parent: "Loader"

    def __init__(self, parent: "Loader", section: str):
        self._parent = parent
        self._section = section

    def get(self, key: str) -> str:
        return self._parent.get(self._section, key)

    def __getitem__(self, key: str) -> str:
        return self.get(key)

    def get_int(self, key: str) -> int:
        return self._parent.get_int(self._section, key)

    def get_float(self, key: str) -> float:
        return self._parent.get_float(self._section, key)

    def get_bool(self, key: str) -> bool:
        return self._parent.get_bool(self._section, key)

    def get_dict(self, key: str) -> Dict[str, Any]:
        return self._parent.get_dict(self._section, key)

    def get_list(self, key: str) -> List[Any]:
        return self._parent.get_list(self._section, key)


class Loader:
    """
    Runtime configuration options, identified by section & key, are sourced in the following
    priority order:

    1. Supplied ``overrides`` dict, ``{"section": {"key": value}}``
    2. Environment variables ``MINIWDL__SECTION__KEY`` (uppercased with double-underscores)
    3. Custom configuration file (mutually exclusive):

       a. filename given to ``__init__``
       b. file named by environment variable ``MINIWDL_CFG``
       c. ``miniwdl.cfg`` in XDG_CONFIG_HOME or XDG_CONFIG_DIRS, usually ``~/.config/miniwdl.cfg``

    4. ``WDL/runtime/config_templates/default.cfg`` from installed package

    If ``filenames`` is an empty list, then no custom configuration file is used. Or ``filenames``
    may be a prioritized list of candidate filenames instead of (a-c) above; the first extant file
    is used, if any.
    """

    _logger: logging.Logger
    _defaults: configparser.ConfigParser
    _options: configparser.ConfigParser
    _overrides: configparser.ConfigParser
    _used: Set[Tuple[str, str]]
    _used_env: Set[str]

    cfg_filename: Optional[str] = None

    def __init__(
        self,
        logger: logging.Logger,
        filenames: Optional[List[str]] = None,
        overrides: Optional[Dict[str, Dict[str, str]]] = None,
    ):
        self._logger = logger
        self._used = set()
        self._used_env = set(["MINIWDL_CFG"])

        common_kws = {}

        self._defaults = configparser.ConfigParser(**common_kws)
        self._options = configparser.ConfigParser(**common_kws)
        self._overrides = configparser.ConfigParser(**common_kws)

        # load default.cfg
        default_cfg = os.path.join(os.path.dirname(__file__), "config_templates", "default.cfg")
        self._logger.debug(_("read configuration defaults", filename=default_cfg))
        self._defaults.read(default_cfg)

        # load miniwdl.cfg, if any
        if filenames is None:
            if "MINIWDL_CFG" in os.environ:
                filenames = os.environ["MINIWDL_CFG"].split(":")
            else:
                filenames = [
                    str(dn.joinpath("miniwdl.cfg"))
                    for dn in reversed(XDG_CONFIG_DIRS + [XDG_CONFIG_HOME])
                ]
        if filenames:
            self._logger.debug(_("searching for configuration files", filenames=filenames))
            filenames = [fn for fn in filenames if os.path.isfile(fn)]
            filenames = filenames[:1]
        if filenames:
            self._logger.info(_("read configuration file", path=filenames[0]))
            self._options.read(filenames)
            self.cfg_filename = filenames[0]
        else:
            self._logger.info("no configuration file found")

        # load overrides
        if overrides:
            self.override(overrides)

    def override(self, options: Dict[str, Dict[str, Any]]) -> None:
        options2 = {}
        for section in options:
            if options[section]:
                options2[section] = {}
                for key in options[section]:
                    v = options[section][key]
                    if isinstance(v, (list, dict, bool)) or v is None:
                        options2[section][key] = json.dumps(v)
                    else:
                        options2[section][key] = str(v)
        if options2:
            self._logger.debug(_("applying configuration overrides", **options2))
            self._overrides.read_dict(options2)

    def get(self, section: str, key: str) -> str:
        section = str(section).lower()
        key = str(key).lower()
        common_kws = {}
        ans = None

        # first check overrides
        if self._overrides.has_option(section, key):
            ans = self._overrides.get(section, key, **common_kws)
        else:
            # ...then environment
            env_key = _env_var_name(section, key)
            if env_key in os.environ:
                ans = os.environ[env_key]
                self._used_env.add(env_key)
            # ...then the config file
            elif self._options.has_option(section, key):
                ans = self._options.get(section, key, **common_kws)
            # ...then the default config
            elif self._defaults.has_option(section, key):
                ans = self._defaults.get(section, key, **common_kws)

        if ans is None:
            if not self.has_section(section):
                raise ConfigMissing(f"missing config section [{section}]")
            raise ConfigMissing(f"missing config option [{section}] {key}")

        self._used.add((section, key))
        ans = _strip(ans)
        if (section, key) not in [("task_runtime", "placeholder_regex")]:
            # expand environment variables unless in a hard-coded list of exceptions
            ans = _expand_env_var(ans)
        return ans

    def has_section(self, section: str) -> bool:
        return (
            self._defaults.has_section(section)
            or self._options.has_section(section)
            or self._overrides.has_section(section)
        )

    def has_option(self, section: str, option: str) -> bool:
        try:
            self.get(section, option)
            return True
        except ConfigMissing:
            return False

    def __getitem__(self, section: str) -> Section:
        return Section(self, section)

    def _parse(self, section: str, key: str, ty: str, parse: Callable[[str], _T]) -> _T:
        ans = self.get(section, key)
        try:
            return parse(ans)
        except:
            self._logger.debug(
                _(
                    "failed to parse configuration option",
                    section=section,
                    key=key,
                    value=ans,
                    expected_type=ty,
                )
            )
            raise ValueError(f"configuration option [{section}] {key} should be {ty}")

    def get_int(self, section: str, key: str) -> int:
        return self._parse(section, key, "int", int)

    def get_float(self, section: str, key: str) -> float:
        return self._parse(section, key, "float", float)

    def get_bool(self, section: str, key: str) -> bool:
        return self._parse(section, key, "bool", _parse_bool)

    def get_dict(self, section: str, key: str) -> Dict[str, Any]:
        return self._parse(section, key, "JSON dict", _parse_dict)

    def get_list(self, section: str, key: str) -> List[Any]:
        return self._parse(section, key, "JSON list", _parse_list)

    def get_all(self, defaults=True):
        options = set()
        for section in self._overrides.sections():
            options |= set((section, key) for key in self._overrides.options(section))
        for section in self._options.sections():
            options |= set((section, key) for key in self._options.options(section))
        for section in self._defaults.sections():
            options |= set((section, key) for key in self._defaults.options(section))

        all = {}
        for (section, key) in options:
            value = self.get(section, key)
            if defaults or value != _expand_env_var(_strip(self._defaults.get(section, key))):
                all[section] = all.get(section, dict())
                all[section][key] = value

        return all

    def log_all(self):
        """
        Write a debug log message with all options
        """
        self._logger.debug(_("configuration", **self.get_all()))

    def log_unused_options(self):
        """
        After a run, log warnings about any options that were set, but were never accessed and
        don't correspond to any default option.
        """
        options = set()
        for section in self._overrides.sections():
            options |= set((section, key) for key in self._overrides.options(section))
        for section in self._options.sections():
            options |= set((section, key) for key in self._options.options(section))
        known_options = set()
        for section in self._defaults.sections():
            known_options |= set((section, key) for key in self._defaults.options(section))
        unused = options - self._used - known_options
        if unused:
            self._logger.warning(
                _("unused configuration", options=[f"{section}:{key}" for (section, key) in unused])
            )

        evs = set(k for k in os.environ if k.upper().startswith("MINIWDL_"))
        ev_unused = evs - self._used_env
        if ev_unused:
            self._logger.warning(_("unused environment", variables=list(ev_unused)))


def _strip(value):
    ans = value
    if ans:
        ans = value.strip()
        if len(ans) >= 2 and (
            (
                ans.startswith("'")
                and ans.endswith("'")
                or (ans.startswith('"') and ans.endswith('"'))
            )
        ):
            ans = ans[1:-1]
    return ans


def _expand_env_var(env_var: str) -> str:
    """
    Expands (potentially nested) env vars by repeatedly applying
    `expandvars` and `expanduser` until interpolation stops having
    any effect.
    """
    interpolated = os.path.expanduser(os.path.expandvars(str(env_var)))
    if interpolated == env_var:
        return interpolated
    return _expand_env_var(interpolated)


def _env_var_name(section: str, key: str) -> str:
    return f"MINIWDL__{section.upper()}__{key.upper()}"


def _parse_bool(v: str) -> bool:
    v = v.lower()
    if v in ("t", "y", "1", "true", "yes"):
        return True
    if v in ("f", "n", "0", "false", "no"):
        return False
    assert False


def _parse_dict(v: str) -> Dict[str, Any]:
    ans = json.loads(v)
    assert isinstance(ans, dict)
    return ans


def _parse_list(v: str) -> List[Any]:
    ans = json.loads(v)
    assert isinstance(ans, list)
    return ans


def default_plugins() -> "Dict[str,List[importlib_metadata.EntryPoint]]":
    import importlib_metadata  # delayed heavy import

    return {
        "file_download": [
            importlib_metadata.EntryPoint(
                group="miniwdl.plugin.file_download",
                name="s3",
                value="WDL.runtime.download:awscli_downloader",
            ),
            importlib_metadata.EntryPoint(
                group="miniwdl.plugin.file_download",
                name="gs",
                value="WDL.runtime.download:gsutil_downloader",
            ),
        ],
        "directory_download": [
            importlib_metadata.EntryPoint(
                group="miniwdl.plugin.directory_download",
                name="s3",
                value="WDL.runtime.download:awscli_directory_downloader",
            )
        ],
        "task": [],
        "workflow": [],
        "container_backend": [
            importlib_metadata.EntryPoint(
                group="miniwdl.plugin.container_backend",
                name="docker_swarm",
                value="WDL.runtime.task_container:SwarmContainer",
            )
        ],
        "cache_backend": [
            importlib_metadata.EntryPoint(
                group="miniwdl.plugin.cache_backend",
                name="dir",
                value="WDL.runtime.cache:CallCache",
            )
        ],
    }


def load_all_plugins(cfg: Loader, group: str) -> Iterable[Tuple[bool, Any]]:
    import importlib_metadata  # delayed heavy import

    defaults = default_plugins()

    assert group in defaults, group
    enable_patterns = cfg["plugins"].get_list("enable_patterns")
    disable_patterns = cfg["plugins"].get_list("disable_patterns")
    for plugin in defaults[group] + list(
        importlib_metadata.entry_points().get(f"miniwdl.plugin.{group}", [])
    ):
        enabled = next(
            (pat for pat in enable_patterns if fnmatchcase(plugin.value, pat)), False
        ) and not next((pat for pat in disable_patterns if fnmatchcase(plugin.value, pat)), False)
        yield (enabled, plugin)


def load_plugins(cfg: Loader, group: str) -> Iterable[Tuple[str, Callable[..., Any]]]:
    yield from (
        (plugin.name, plugin.load()) for enabled, plugin in load_all_plugins(cfg, group) if enabled
    )



import glob
import json
import logging
import os
import stat
import random
import shutil
import tempfile
import time
import unittest
import subprocess
from unittest.mock import MagicMock, patch

from .context import WDL
from WDL import values_from_json, values_to_json
from WDL.runtime.cache import CallCache


class TestCallCache(unittest.TestCase):
    test_wdl: str = R"""
        version 1.0
        task hello_blank {
            # comment1
            input  {
                String who     # comment2
                Array[String]? what

                Map[String,Map[String,String]]? where
            }     


            command <<<
                # comment3
                echo "Hello, ~{who}!"
            >>>
            output {#comment4
                Int count = 12

            }
        }#comment5
        """
    ordered_input_dict = {
        "what": ["a", "ab", "b", "bc"],
        "where": {"places": {"Minneapolis": "a", "SanFan": "b"}},
        "who": "Alyssa",
    }
    doc = WDL.parse_document(test_wdl)
    cache_dir = '/tmp/cache/'
    struct_task: str = R"""
            version 1.0
            struct Box {
                Array[File] str
            }
            task hello {
                input {
                    Box box
                }
                command {
                    echo "Hello, world!"
                }
                output {
                    Int count = 13
                }
            }
            """
    test_wdl_with_output_files: str = R"""
        version 1.0
        task hello {
            String who
            File foo = write_lines(["foo","bar","baz"])
            File tsv = write_tsv([["one", "two", "three"], ["un", "deux", "trois"]])
            File json = write_json({"key1": "value1", "key2": "value2"})

            command <<<
                echo "Hello, ~{who}!"
            >>>

            output {
                File o_json = json
                File a_tsv = tsv
                File whynot = write_lines(["foo","bar","baz"])
                Int count = 13
                String ans = stdout()
            }
        }
        """

    @classmethod
    def setUpClass(cls):
        logging.basicConfig(level=logging.DEBUG, format='%(name)s %(levelname)s %(message)s')
        cls.logger = logging.getLogger(cls.__name__)
        cls.cfg = WDL.runtime.config.Loader(cls.logger, [])
        cls.cfg.override(
            {"call_cache": {
                "put": True,
                "get": True,
                "dir": cls.cache_dir
            }
        })

    def setUp(self):
        """
        initialize docker & provision temporary directory for a test (self._dir)
        """
        self._dir = tempfile.mkdtemp(prefix=f"miniwdl_test_{self.id()}_")

    def tearDown(self):
        shutil.rmtree(self._dir)
        try:
            shutil.rmtree(self.cache_dir)
        except FileNotFoundError:
            print("No cache directory to delete")

    def _run(self, wdl: str, inputs=None, expected_exception: Exception = None, cfg=None):
        """
        run workflow/task & return outputs dict
        """
        logger = logging.getLogger(self.id())
        cfg = cfg or WDL.runtime.config.Loader(logger, [])
        try:
            with tempfile.NamedTemporaryFile(dir=self._dir, suffix=".wdl", delete=False) as outfile:
                outfile.write(wdl.encode("utf-8"))
                wdlfn = outfile.name
            doc = WDL.load(wdlfn)
            target = doc.workflow or doc.tasks[0]
            if isinstance(inputs, dict):
                inputs = WDL.values_from_json(inputs, target.available_inputs, target.required_inputs)
            rundir, outputs = WDL.runtime.run(cfg, target, (inputs or WDL.Env.Bindings()), run_dir=self._dir)

            self._rundir = rundir
        except Exception as exn:
            while isinstance(exn, WDL.runtime.RunFailed):
                exn = exn.__context__
            if expected_exception:
                self.assertIsInstance(exn, expected_exception)
                return exn
            raise
        self.assertIsNone(expected_exception, str(expected_exception) + " not raised")

        return rundir, outputs

    def test_input_digest_sorts_keys(self):
        # Note this fails if input array is reordered

        ordered_inputs = values_from_json(
            self.ordered_input_dict, self.doc.tasks[0].available_inputs)
        unordered_inputs = values_from_json(
            {
                "where": {"places": {"SanFan": "b", "Minneapolis": "a"}},
                "what": ["a", "ab", "b", "bc"],
                "who": "Alyssa"
            }, self.doc.tasks[0].available_inputs)

        ordered_digest = WDL.Value.digest_env(ordered_inputs)
        unordered_digest = WDL.Value.digest_env(unordered_inputs)
        self.assertEqual(ordered_digest, unordered_digest)

    def test_normalization(self):
        desc = self.doc.tasks[0]._digest_source()
        self.assertEqual(desc, R"""
version 1.0
task hello_blank {
input  {
String who
Array[String]? what
Map[String,Map[String,String]]? where
}
            command <<<
                # comment3
                echo "Hello, ~{who}!"
            >>>
output {
Int count = 12
}
}
        """.strip())

    def test_task_input_cache_matches_output(self):
        # run task, check output matches what was stored in run_dir
        cache = CallCache(cfg=self.cfg, logger=self.logger)
        rundir, outputs = self._run(self.test_wdl, self.ordered_input_dict, cfg=self.cfg)
        inputs = values_from_json(
            self.ordered_input_dict, self.doc.tasks[0].available_inputs)
        input_digest = WDL.Value.digest_env(inputs)
        task_digest = self.doc.tasks[0].digest
        with open(os.path.join(self.cache_dir, f"{self.doc.tasks[0].name}/{task_digest}/{input_digest}.json")) as f:
            read_data = json.loads(f.read())
        self.assertEqual(read_data, WDL.values_to_json(outputs))

    def test_cache_prevents_task_rerun(self):
        # run task twice, check _try_task not called for second run

        mock = MagicMock(side_effect=WDL.runtime.task._try_task)

        # test mock is called
        with patch('WDL.runtime.task._try_task', mock):
            self._run(self.test_wdl, self.ordered_input_dict, cfg=self.cfg)
        self.assertEqual(mock.call_count, 1)

        # call real _try_task function
        self._run(self.test_wdl, self.ordered_input_dict, cfg=self.cfg)

        # test mock is not called once cache is available
        new_mock = MagicMock(side_effect=WDL.runtime.task._try_task)

        with patch('WDL.runtime.task._try_task', new_mock):
            self._run(self.test_wdl, self.ordered_input_dict, cfg=self.cfg)

        self.assertEqual(new_mock.call_count, 0)

    def test_default_config_does_not_use_cache(self):
        # run task twice, check _try_task called for second run
        mock = MagicMock(side_effect=WDL.runtime.task._try_task)

        # test mock is called
        with patch('WDL.runtime.task._try_task', mock):
            self._run(self.test_wdl, self.ordered_input_dict)
        self.assertEqual(mock.call_count, 1)

        # call real _try_task function
        self._run(self.test_wdl, self.ordered_input_dict)

        # test mock is not called once cache is available
        new_mock = MagicMock(side_effect=WDL.runtime.task._try_task)

        with patch('WDL.runtime.task._try_task', new_mock):
            self._run(self.test_wdl, self.ordered_input_dict)

        self.assertEqual(new_mock.call_count, 1)

    def test_get_cache_return_value_matches_outputs(self):
        cache = CallCache(cfg=self.cfg, logger=self.logger)
        rundir, outputs = self._run(self.test_wdl, self.ordered_input_dict, cfg=self.cfg)
        inputs = values_from_json(
            self.ordered_input_dict, self.doc.tasks[0].available_inputs)
        input_digest = WDL.Value.digest_env(inputs)
        task_digest = self.doc.tasks[0].digest
        cache_value = cache.get(key=f"{self.doc.tasks[0].name}/{task_digest}/{input_digest}",
                                output_types=self.doc.tasks[0].effective_outputs,
                                inputs=inputs)
        self.assertEqual(values_to_json(outputs), values_to_json(cache_value))

    def test_a_task_with_the_same_inputs_and_different_commands_doesnt_pull_from_the_cache(self):
        # run task twice, once with original wdl, once with updated wdl command, check _try_task  called for second run
        new_test_wdl: str = R"""
               version 1.0
               task hello_blank {
                   input {
                       String who
                       Array[String]? what
                       Map[String,Map[String,String]]? where
                   }
                   command <<<
                       echo "Heyyyyy, ~{who}!"
                   >>>
                   output {
                       Int count = 12
                   }
               }
               """

        #  _try_task function for original wdl
        self._run(self.test_wdl, self.ordered_input_dict, cfg=self.cfg)

        # test _try_task is called when task def changes (with same inputs)
        mock = MagicMock(side_effect=WDL.runtime.task._try_task)

        with patch('WDL.runtime.task._try_task', mock):
            self._run(new_test_wdl, self.ordered_input_dict, cfg=self.cfg)

        self.assertEqual(mock.call_count, 1)

    def test_a_task_with_the_same_inputs_and_different_outputs_doesnt_pull_from_the_cache(self):
        # run task twice, once with original wdl, once with updated wdl command, check _try_task  called for second run
        new_test_wdl: str = R"""
                  version 1.0
                  task hello_blank {
                      input {
                          String who
                          Array[String]? what
                          Map[String,Map[String,String]]? where
                      }
                      command <<<
                          echo "Hello, ~{who}!"
                      >>>
                      output {
                          Int count = 13
                      }
                  }
                  """

        #  _try_task function for original wdl
        self._run(self.test_wdl, self.ordered_input_dict, cfg=self.cfg)

        # test _try_task is called when task def changes (with same inputs)
        mock = MagicMock(side_effect=WDL.runtime.task._try_task)

        with patch('WDL.runtime.task._try_task', mock):
            self._run(new_test_wdl, self.ordered_input_dict, cfg=self.cfg)

        self.assertEqual(mock.call_count, 1)

    def test_struct_handling(self):
        with open(os.path.join(self._dir, "randomFile.txt"), "w") as outfile:
            outfile.write("Gotta put something here")
        inputs = {"box": {"str": [os.path.join(self._dir, "randomFile.txt")]}}
        mock = MagicMock(side_effect=WDL.runtime.task._try_task)
        # test mock is called
        with patch('WDL.runtime.task._try_task', mock):
            self._run(self.struct_task, inputs, cfg=self.cfg)
        self.assertEqual(mock.call_count, 1)
        # run for real
        self._run(self.struct_task, inputs, cfg=self.cfg)

        new_mock = MagicMock(side_effect=WDL.runtime.task._try_task)

        # test mock not called for cached tasks containing a struct
        with patch('WDL.runtime.task._try_task', new_mock):
            self._run(self.struct_task, inputs, cfg=self.cfg)
        self.assertEqual(new_mock.call_count, 0)

    def test_cache_not_used_when_output_files_deleted(self):
        inputs = {"who": "Alyssa"}
        self._run(self.test_wdl_with_output_files, inputs, cfg=self.cfg)
        # test mock is not called once cache is available
        mock = MagicMock(side_effect=WDL.runtime.task._try_task)

        with patch('WDL.runtime.task._try_task', mock):
            self._run(self.test_wdl_with_output_files, inputs, cfg=self.cfg)

        self.assertEqual(mock.call_count, 0)

        # delete files
        for x in glob.glob(f"{self._dir}/*_hello/out/a_tsv"):
            shutil.rmtree(x)

        # test mock is called now that cached file has been deleted
        with patch('WDL.runtime.task._try_task', mock):
            self._run(self.test_wdl_with_output_files, inputs, cfg=self.cfg)
        self.assertEqual(mock.call_count, 1)

    def test_cache_not_used_when_output_files_updated_after_cache_creation(self):
        inputs = {"who": "Bethie"}
        self._run(self.test_wdl_with_output_files, inputs, cfg=self.cfg)
        # change modified time on outputs
        time.sleep(0.1)
        for x in glob.glob(f"{self._dir}/*_hello/out/a_tsv/*"):
            os.utime(x)

        # check that mock is called now that output file is older than cache file
        mock = MagicMock(side_effect=WDL.runtime.task._try_task)
        with patch('WDL.runtime.task._try_task', mock):

            self._run(self.test_wdl_with_output_files, inputs, cfg=self.cfg)
        self.assertEqual(mock.call_count, 1)

    def test_cache_not_used_when_output_files_but_not__sym_links_updated_after_cache_creation(self):
        inputs = {"who": "Bethie"}
        self._run(self.test_wdl_with_output_files, inputs, cfg=self.cfg)
        # change modified time on outputs
        time.sleep(0.1)
        for x in glob.glob(f"{self._dir}/*_hello/out/a_tsv/*"):
            os.utime(x, follow_symlinks=False)

        # check that mock is called now that output file is older than cache file
        mock = MagicMock(side_effect=WDL.runtime.task._try_task)
        with patch('WDL.runtime.task._try_task', mock):
            self._run(self.test_wdl_with_output_files, inputs, cfg=self.cfg)
        self.assertEqual(mock.call_count, 1)

    def test_cache_not_used_when_file_in_array_recently_updated(self):
        filenames = ["file1", "file2", "file3", "butterfinger"]
        inputs = {"files": []}
        for fn in filenames:
            fn = os.path.join(self._dir, fn)
            with open(fn, "w") as outfile:
                print(fn, file=outfile)
            inputs["files"].append(fn)

        wdl = """
                version 1.0
                task return_file_array {
                    input {
                        Array[File] files
                    }
                    command <<<
                        set -euxo pipefail
                        mkdir files_out
                        find _miniwdl_inputs -type f -print0 | xargs -0 -iXXX cp XXX files_out/
                    >>>
                    output {
                        Array[File] files_out = glob("files_out/*")
                    }
                }
                """

        self._run(wdl, inputs, cfg=self.cfg)
        #check cache used
        mock = MagicMock(side_effect=WDL.runtime.task._try_task)
        with patch('WDL.runtime.task._try_task', mock):
            self._run(wdl, inputs, cfg=self.cfg)
        self.assertEqual(mock.call_count, 0)
        # change time
        time.sleep(0.1)
        for x in glob.glob(f"{self._dir}/*_return_file_array/work/files_out/file1"):
            os.utime(x)
        # check cache not used
        with patch('WDL.runtime.task._try_task', mock):
            self._run(wdl, inputs, cfg=self.cfg)
        self.assertEqual(mock.call_count, 1)

    def test_cache_not_used_when_input_file_recently_updated(self):
        filenames = ["file1", "file2", "file3", "butterfinger"]
        inputs = {"files": []}
        for fn in filenames:
            fn = os.path.join(self._dir, fn)
            with open(fn, "w") as outfile:
                print(fn, file=outfile)
            inputs["files"].append(fn)

        wdl = """
                version 1.0
                task return_file_array {
                    input {
                        Array[File] files
                    }
                    command <<<
                        echo "Hello"
                    >>>
                    output {
                         Int count = 13
                    }
                }
                """

        self._run(wdl, inputs, cfg=self.cfg)
        #check cache used
        mock = MagicMock(side_effect=WDL.runtime.task._try_task)
        with patch('WDL.runtime.task._try_task', mock):
            self._run(wdl, inputs, cfg=self.cfg)
        self.assertEqual(mock.call_count, 0)
        # change time on input file
        time.sleep(0.1)
        for x in glob.glob(f"{self._dir}/butterfinger"):
            os.utime(x)
        # check cache not used
        with patch('WDL.runtime.task._try_task', mock):
            self._run(wdl, inputs, cfg=self.cfg)
        self.assertEqual(mock.call_count, 1)

    def test_directory_coherence(self):
        # test outputting files/subdirectories inside input Directory
        wdl = R"""
        version development
        task t {
            input {
                Directory d
            }
            command {}
            output {
                Array[File] files = ["~{d}/alice.txt", "~{d}/sub/bob.txt"]
                Array[Directory] dirs = ["~{d}/sub/dir"]
            }
        }
        """
        os.makedirs(os.path.join(self._dir, "d/sub/dir"))
        with open(os.path.join(self._dir, "d/alice.txt"), mode="w") as outfile:
            print("Alice", file=outfile)
        with open(os.path.join(self._dir, "d/sub/bob.txt"), mode="w") as outfile:
            print("Bob", file=outfile)
        with open(os.path.join(self._dir, "d/sub/dir/carol.txt"), mode="w") as outfile:
            print("Carol", file=outfile)
        inp = {"d": os.path.join(self._dir, "d")}
        outp = self._run(wdl, inp, cfg=self.cfg)

        WDL.Value.rewrite_env_files(outp[1], lambda fn: fn)  # game coverage of deprecated fn

        mock = MagicMock(side_effect=WDL.runtime.task._try_task)
        with patch('WDL.runtime.task._try_task', mock):
            # control
            self._run(wdl, inp, cfg=self.cfg)
            self.assertEqual(mock.call_count, 0)

            # touch a file & check cache invalidated
            subprocess.run(["touch", os.path.join(self._dir, "d/sub/dir/carol.txt")], check=True)
            self._run(wdl, inp, cfg=self.cfg)
            self.assertEqual(mock.call_count, 1)

            # add a symlink
            time.sleep(0.1)
            os.symlink("sub/dir", os.path.join(self._dir, "d/link1"))
            self._run(wdl, inp, cfg=self.cfg)
            self.assertEqual(mock.call_count, 2)

            # delete the symlink
            time.sleep(0.1)
            os.unlink(os.path.join(self._dir, "d/link1"))
            self._run(wdl, inp, cfg=self.cfg)
            self.assertEqual(mock.call_count, 3)

            # control
            self._run(wdl, inp, cfg=self.cfg)
            self.assertEqual(mock.call_count, 3)

    test_workflow_wdl = R"""
    version development

    struct Person {
        String first
        String? middle
        String last
    }

    workflow multihello {
        input {
            Array[File] people_json
        }
        # COMMENT
        scatter (person_json in people_json) {
            call read_person {
                input:
                json = person_json
            }
            call hello {
                input:
                full_name = read_person.full_name
            }
        }

        output {
            Array[File] messages = hello.message
        }
    }

    task read_person {
        input {
            File json
        }

        Person person = read_json(json)

        command {}

        output {
            File full_name = write_lines([sep(" ", [person.first, person.last])])
        }
    }

    task hello {
        input {
            File full_name
            String? greeting = "Hello"
        }

        command <<<
            echo '~{greeting}, ~{read_string(full_name)}!'
        >>>

        output {
            File message = stdout()
        }
    }

    task uncalled {
        input {
            Int i = 0
            Person? p
        }
        command {}
    }
    """

    def test_workflow_digest(self):
        doc = WDL.parse_document(self.test_workflow_wdl)
        doc.typecheck()

        # ensure digest is sensitive to changes in the struct type and called task (but not the
        # uncalled task, or comments/whitespace)
        doc2 = WDL.parse_document(self.test_workflow_wdl.replace("String? middle", ""))
        doc2.typecheck()
        self.assertNotEqual(doc.workflow.digest, doc2.workflow.digest)

        doc2 = WDL.parse_document(self.test_workflow_wdl.replace('"Hello"', '"Hi"'))
        doc2.typecheck()
        self.assertNotEqual(doc.workflow.digest, doc2.workflow.digest)

        doc2 = WDL.parse_document(self.test_workflow_wdl.replace('i = 0', 'i = 1'))
        doc2.typecheck()
        self.assertEqual(doc.workflow.digest, doc2.workflow.digest)

        doc2 = WDL.parse_document(self.test_workflow_wdl.replace('# COMMENT', '#'))
        doc2.typecheck()
        self.assertEqual(doc.workflow.digest, doc2.workflow.digest)

        doc2 = WDL.parse_document(self.test_workflow_wdl.replace('# COMMENT', '\n\n'))
        doc2.typecheck()
        self.assertEqual(doc.workflow.digest, doc2.workflow.digest)

    def test_workflow_cache(self):
        with open(os.path.join(self._dir, "alyssa.json"), mode="w") as outfile:
            print('{"first":"Alyssa","last":"Hacker"}', file=outfile)
        with open(os.path.join(self._dir, "ben.json"), mode="w") as outfile:
            print('{"first":"Ben","last":"Bitdiddle"}', file=outfile)
        inp = {"people_json": [os.path.join(self._dir, "alyssa.json"), os.path.join(self._dir, "ben.json")]}
        rundir1, outp = self._run(self.test_workflow_wdl, inp, cfg=self.cfg)

        wmock = MagicMock(side_effect=WDL.runtime.workflow._workflow_main_loop)
        tmock = MagicMock(side_effect=WDL.runtime.task._try_task)
        with patch('WDL.runtime.workflow._workflow_main_loop', wmock), patch('WDL.runtime.task._try_task', tmock):
            # control
            rundir2, outp2 = self._run(self.test_workflow_wdl, inp, cfg=self.cfg)
            self.assertEqual(wmock.call_count, 0)
            self.assertEqual(tmock.call_count, 0)
            outp_inodes = set()
            WDL.Value.rewrite_env_paths(outp, lambda p: outp_inodes.add(os.stat(p.value)[stat.ST_INO]))
            outp2_inodes = set()
            WDL.Value.rewrite_env_paths(outp2, lambda p: outp2_inodes.add(os.stat(p.value)[stat.ST_INO]))
            self.assertEqual(outp_inodes, outp2_inodes)

            with open(os.path.join(rundir1, "outputs.json")) as outputs1:
                with open(os.path.join(rundir2, "outputs.json")) as outputs2:
                    assert outputs1.read() == outputs2.read()

            # touch a file & check cache invalidated
            with open(os.path.join(self._dir, "alyssa.json"), mode="w") as outfile:
                print('{"first":"Alyssa","last":"Hacker","middle":"P"}', file=outfile)
            _, outp2 = self._run(self.test_workflow_wdl, inp, cfg=self.cfg)
            self.assertEqual(wmock.call_count, 1)
            self.assertEqual(tmock.call_count, 2)  # reran Alyssa, cached Ben
            self.assertNotEqual(WDL.values_to_json(outp), WDL.values_to_json(outp2))



from setuptools import setup, find_packages

setup(
    name='miniwdl_task_omnibus_example',
    version='0.0.1',
    description='miniwdl task runtime plugin (omnibus example)',
    author='Wid L. Hacker',
    py_modules=["miniwdl_task_omnibus_example"],
    python_requires='>=3.6',
    setup_requires=['reentry'],
    install_requires=["boto3"],
    reentry_register=True,
    entry_points={
        'miniwdl.plugin.task': ['omnibus = miniwdl_task_omnibus_example:main'],
    }
)



import unittest, inspect, subprocess, tempfile, os, glob, json, urllib, urllib.request
from .context import WDL
import WDL.Lint


class Lint(unittest.TestCase):
    # test programmatic access to linter results
    # this code should be kept in sync with the example shown in the WDL/Lint.py docstring.
    def test_api(self):
        doc = WDL.load("ZarrUtils.wdl", path=["test_corpi/HumanCellAtlas/skylab/library/tasks"])
        WDL.Lint._shellcheck_available = False
        lint = WDL.Lint.collect(WDL.Lint.lint(doc, descend_imports=False))
        for (pos, lint_class, message, suppressed) in lint:
            assert isinstance(pos, WDL.SourcePosition)
            assert isinstance(lint_class, str) and isinstance(message, str)
            if not suppressed:
                print(
                    json.dumps(
                        {
                            "uri": pos.uri,
                            "abspath": pos.abspath,
                            "line": pos.line,
                            "end_line": pos.end_line,
                            "column": pos.column,
                            "end_column": pos.end_column,
                            "lint": lint_class,
                            "message": message,
                        }
                    )
                )
        self.assertEqual(len(lint), 2)


async def read_source(uri, path, importer_uri):
    if uri.startswith("http:") or uri.startswith("https:"):
        # Note: we should permit web imports only in corpi which are careful to pin a specific and
        # highly-available revision
        fn = os.path.join(tempfile.mkdtemp(prefix="miniwdl_import_uri_"), os.path.basename(uri))
        urllib.request.urlretrieve(uri, filename=fn)
        with open(fn, "r") as infile:
            return WDL.ReadSourceResult(infile.read(), os.path.abspath(fn))
    return await WDL.read_source_default(uri, path, importer_uri)


def wdl_corpus(dir, path=[], blocklist=[], expected_lint={}, check_quant=True):
    def decorator(test_klass):

        test_klass._lint_count = {}
        test_klass._expected_lint = expected_lint
        test_klass.tearDownClass = classmethod(check_lint)

        prefix = test_klass.__name__
        files = glob.glob(os.path.join(*(dir + ["*.wdl"])), recursive=True)
        assert (
            len(files) > 0
        ), "{} test corpus missing from {}; please `git submodule update --init --recursive`".format(
            prefix, os.path.join(*dir)
        )
        gpath = []
        for p in path:
            gpath = gpath + glob.glob(os.path.join(*p), recursive=True)
        for fn in files:
            name = os.path.split(fn)[1]
            name = name[:-4]
            if name not in blocklist:
                name = "test_" + prefix + "_" + name.replace(".", "_")
                while hasattr(test_klass, name):
                    name += "_"

                def t(self, fn=fn):
                    # load & lint the document to verify the lint count
                    try:
                        doc = WDL.load(
                            fn, path=gpath, check_quant=check_quant, read_source=read_source
                        )
                    except Exception as exn:
                        if isinstance(exn, WDL.Error.MultipleValidationErrors):
                            for subexn in exn.exceptions:
                                print(subexn.node.pos)
                        if hasattr(exn, "node"):
                            print(exn.node.pos)
                        raise
                    WDL.Lint.lint(doc)
                    for _, linter, _, suppressed in WDL.Lint.collect(doc):
                        test_klass._lint_count[linter] = 1 + test_klass._lint_count.get(linter, 0)
                        if suppressed:
                            test_klass._lint_count[
                                "_suppressions"
                            ] = 1 + test_klass._lint_count.get("_suppressions", 0)
                    print("\n" + os.path.basename(fn))
                    WDL.CLI.outline(doc, 0, show_called=(doc.workflow is not None))
                    WDL.copy_source(
                        doc, tempfile.mkdtemp(prefix=f"miniwdl_test_copy_source_{prefix}")
                    )

                    if doc.workflow:
                        validate_workflow_graph(doc.workflow)

                    # also attempt load with the opposite value of check_quant,
                    # exercising additional code paths
                    try:
                        doc = WDL.load(
                            fn, path=gpath, check_quant=not check_quant, read_source=read_source
                        )
                    except (
                        WDL.Error.ImportError,
                        WDL.Error.ValidationError,
                        WDL.Error.MultipleValidationErrors,
                    ):
                        pass

                setattr(test_klass, name, t)

        return test_klass

    return decorator


def check_lint(cls):
    if "CommandShellCheck" in cls._lint_count:
        # because we don't control installed shellcheck version
        del cls._lint_count["CommandShellCheck"]
    if cls._lint_count != cls._expected_lint:
        raise Exception(
            "Lint results changed for {}; expected: {} got: {}".format(
                cls.__name__, str(cls._expected_lint), str(cls._lint_count)
            )
        )


def validate_workflow_graph(workflow):
    def visit_section(body, outside_nodes):
        body_nodes = set()
        body_gathers = set()
        for node in body:
            body_nodes.add(node.workflow_node_id)
            if isinstance(node, WDL.WorkflowSection):
                for g in node.gathers.values():
                    body_gathers.add(g.workflow_node_id)
        assert not (body_nodes & body_gathers)
        assert not (outside_nodes & (body_nodes | body_gathers))
        for node in body:
            unk = node.workflow_node_dependencies - (outside_nodes | body_nodes | body_gathers)
            assert not unk, str((node.workflow_node_id, unk))
            if isinstance(node, WDL.WorkflowSection):
                visit_section(
                    node.body, (outside_nodes | body_nodes | body_gathers) - set(node.gathers)
                )

    visit_section(workflow.body, set(inp.workflow_node_id for inp in workflow.inputs or []))


@wdl_corpus(
    ["test_corpi/HumanCellAtlas/skylab/library/tasks/**"],
    expected_lint={"UnusedDeclaration": 2, "UnknownRuntimeKey": 1, "MissingVersion": 26},
)
class HCAskylab_task(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/HumanCellAtlas/skylab/pipelines/**"],
    path=[["test_corpi/HumanCellAtlas/skylab/library/tasks"]],
    expected_lint={
        "UnusedDeclaration": 12,
        "NameCollision": 3,
        "UnknownRuntimeKey": 3,
        "StringCoercion": 1,
        "MixedIndentation": 1,
        "FileCoercion": 1,
        "MissingVersion": 34,
        "UnnecessaryQuantifier": 3,
    },
)
class HCAskylab_workflow(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/gatk-workflows/five-dollar-genome-analysis-pipeline/**"],
    expected_lint={"UnusedDeclaration": 5, "NameCollision": 2, "UnusedImport": 8},
)
class GATK_five_dollar(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/gatk-workflows/gatk4-germline-snps-indels/**"],
    expected_lint={
        "UnusedDeclaration": 3,
        "StringCoercion": 20,
        "FileCoercion": 3,
        "UnknownRuntimeKey": 1,
        "MissingVersion": 4,
    },
)
class gatk4_germline_snps_indels(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/gatk-workflows/gatk4-somatic-snvs-indels/**"],
    expected_lint={
        "OptionalCoercion": 50,
        "NonemptyCoercion": 4,
        "UnusedDeclaration": 29,
        "ForwardReference": 6,
        "StringCoercion": 20,
        "MixedIndentation": 6,
        "MissingVersion": 6,
    },
    check_quant=False,
)
class gatk4_somatic_snvs_indels(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/gatk-workflows/gatk4-cnn-variant-filter/**"],
    expected_lint={
        "UnusedDeclaration": 21,
        "OptionalCoercion": 24,
        "FileCoercion": 8,
        "UnusedCall": 1,
        "StringCoercion": 4,
        "MissingVersion": 6,
    },
    check_quant=False,
)
class gatk4_cnn_variant_filter(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/gatk-workflows/broad-prod-wgs-germline-snps-indels/**"],
    blocklist=["JointGenotypingWf"],
    expected_lint={
        "StringCoercion": 50,
        "UnusedDeclaration": 10,
        "ArrayCoercion": 4,
        "MissingVersion": 2,
    },
    check_quant=False,
)
class broad_prod_wgs(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/broadinstitute/gtex-pipeline/**"],
    # need URI import
    blocklist=["rnaseq_pipeline_bam", "rnaseq_pipeline_fastq"],
    expected_lint={
        "IncompleteCall": 30,
        "UnusedDeclaration": 3,
        "NameCollision": 4,
        "MissingVersion": 23,
    },
)
class GTEx(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/DataBiosphere/topmed-workflows/**"],
    # need URI import
    blocklist=[
        "CRAM_md5sum_checker_wrapper",
        "checker-workflow-wrapping-alignment-workflow",
        "topmed_freeze3_calling",
        "topmed_freeze3_calling_checker",
        "u_of_michigan_aligner_checker",
    ],
    expected_lint={
        "StringCoercion": 27,
        "UnusedDeclaration": 74,
        "OptionalCoercion": 1,
        "MissingVersion": 8,
        "UnnecessaryQuantifier": 1,
    },
    check_quant=False,
)
class TOPMed(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/broadinstitute/viral-ngs/pipes/WDL/workflows"],
    path=[["test_corpi/broadinstitute/viral-ngs/pipes/WDL/workflows/tasks"]],
    expected_lint={
        "UnusedDeclaration": 23,
        "NameCollision": 9,
        "IncompleteCall": 44,
        "UnusedImport": 1,
        "SelectArray": 4,
        "MissingVersion": 62,
        "UnnecessaryQuantifier": 191,
    },
)
class ViralNGS(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/ENCODE-DCC/chip-seq-pipeline2/**"],
    expected_lint={
        "StringCoercion": 208,
        "FileCoercion": 154,
        "NameCollision": 16,
        "OptionalCoercion": 64,
        "MixedIndentation": 32,
        "MissingVersion": 31,
    },
    check_quant=False,
)
class ENCODE_ChIPseq(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/ENCODE-DCC/atac-seq-pipeline/**"],
    expected_lint={
        "UnusedDeclaration": 63,
        "MixedIndentation": 15,
        "OptionalCoercion": 1020,
        "UnusedCall": 45,
        "StringCoercion": 30,
        "FileCoercion": 71,
        "MissingVersion": 29,
    },
    check_quant=False,
)
class ENCODE_ATACseq(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/ENCODE-DCC/rna-seq-pipeline/**"],
    expected_lint={
        "StringCoercion": 2,
        "UnusedDeclaration": 3,
        "IncompleteCall": 3,
        "MissingVersion": 3,
    },
)
class ENCODE_RNAseq(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/ENCODE-DCC/wgbs-pipeline/**"],
    expected_lint={
        "StringCoercion": 9,
        "UnusedDeclaration": 1,
        "MixedIndentation": 1,
        "MissingVersion": 5,
    },
)
class ENCODE_WGBS(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/dnanexus/dxWDL/test/**"],
    blocklist=[
        # output/call name collision (draft-2)
        "conditionals2",
        # decl/output name collision
        "two_levels",
        # use dnanexus extensions
        "call_native",
        "call_native_app",
        "call_native_v1",
    ],
    path=[["test_corpi/dnanexus/dxWDL/test/imports/lib"]],
    expected_lint={
        "UnusedDeclaration": 34,
        "UnusedCall": 16,
        "NameCollision": 2,
        "OptionalCoercion": 3,
        "FileCoercion": 3,
        "StringCoercion": 2,
        "UnnecessaryQuantifier": 1,
        "MissingVersion": 52,
        "UnnecessaryQuantifier": 10,
    },
    check_quant=False,
)
class dxWDL(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/contrived/**"],
    expected_lint={
        "_suppressions": 8,
        "UnusedImport": 4,
        "NameCollision": 27,
        "StringCoercion": 6,
        "FileCoercion": 3,
        "NonemptyCoercion": 1,
        "UnnecessaryQuantifier": 5,
        "UnusedDeclaration": 2,
        "IncompleteCall": 2,
        "SelectArray": 1,
        "MissingVersion": 7,
        "UnboundDeclaration": 1,
        "UnverifiedStruct": 3,
        "Deprecated": 2,
    },
    blocklist=["check_quant", "incomplete_call"],
)
class Contrived(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/contrived/**"],
    expected_lint={
        "_suppressions": 16,
        "UnusedImport": 6,
        "NameCollision": 43,
        "StringCoercion": 11,
        "FileCoercion": 5,
        "OptionalCoercion": 3,
        "NonemptyCoercion": 2,
        "UnnecessaryQuantifier": 9,
        "UnusedDeclaration": 9,
        "IncompleteCall": 3,
        "ArrayCoercion": 2,
        "SelectArray": 4,
        "MissingVersion": 11,
        "UnboundDeclaration": 1,
        "UnverifiedStruct": 3,
        "Deprecated": 2,
    },
    check_quant=False,
    blocklist=["incomplete_call"],
)
class Contrived2(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/biowdl/tasks/**"],
    blocklist=[
        # these use the pattern 'input { Type? x = default }' and need check_quant=False
        "mergecounts",
        "somaticseq",
    ],
    expected_lint={
        "OptionalCoercion": 9,
        "UnusedDeclaration": 18,
        "NonemptyCoercion": 1,
        "NameCollision": 1,
        "SelectArray": 1,
        "UnverifiedStruct": 1,
        "UnnecessaryQuantifier": 8,
    },
)
class BioWDLTasks(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/biowdl/aligning/**"],
    expected_lint={
        "OptionalCoercion": 11,
        "UnusedDeclaration": 12,
        "NonemptyCoercion": 1,
        "NameCollision": 1,
        "UnverifiedStruct": 1,
        "UnnecessaryQuantifier": 13,
    },
    check_quant=False,
)
class BioWDLAligning(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/biowdl/expression-quantification/**"],
    expected_lint={
        "OptionalCoercion": 11,
        "UnusedDeclaration": 12,
        "NonemptyCoercion": 3,
        "NameCollision": 1,
        "UnverifiedStruct": 1,
        "UnnecessaryQuantifier": 9,
    },
    check_quant=False,
)
class BioWDLExpressionQuantification(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/biowdl/somatic-variantcalling"],
    expected_lint={
        "UnusedImport": 2,
        "OptionalCoercion": 11,
        "UnusedDeclaration": 11,
        "NonemptyCoercion": 37,
        "SelectArray": 5,
        "UnnecessaryQuantifier": 3,
    },
    check_quant=False,
)
class BioWDLSomaticVariantCalling(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/biowdl/small-rna"],
    expected_lint={
        "UnusedDeclaration": 8,
        "SelectArray": 2,
        "NonemptyCoercion": 3,
        "UnusedCall": 1,
        "UnverifiedStruct": 1,
        "UnnecessaryQuantifier": 7,
    },
    check_quant=False,
)
class BioWDLSmallRNA(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/broadinstitute/warp/pipelines/broad/**"],
    path=[["test_corpi/broadinstitute/warp/tasks"]],
    expected_lint={
        "UnusedImport": 22,
        "UnusedCall": 1,
        "StringCoercion": 86,
        "UnusedDeclaration": 106,
        "NameCollision": 12,
        "ForwardReference": 5,
        "NonemptyCoercion": 4,
        "FileCoercion": 17,
    },
)
class warp_pipelines_broad(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/broadinstitute/warp/pipelines/cemba/**"],
    expected_lint={
        "UnusedDeclaration": 1,
    },
)
class warp_pipelines_cemba(unittest.TestCase):
    pass


@wdl_corpus(
    ["test_corpi/broadinstitute/warp/pipelines/skylab/**"],
    expected_lint={
        "UnusedDeclaration": 3,
        "UnnecessaryQuantifier": 3,
        "StringCoercion": 4,
        "FileCoercion": 3,
        "NameCollision": 3,
    },
)
class warp_pipelines_skylab(unittest.TestCase):
    pass



import unittest, inspect, tempfile, os, pickle
from typing import Optional
from .context import WDL

class TestTasks(unittest.TestCase):
    def test_wc(self):
        variants = [
        """
        task wc {
            input {
                String in
            }
            command {
                echo "~{in}" | wc
            }
            output {
                String ans = stdout()
            }
        }
        """,
        """
        task wc {
            input {
                String in
            }
            command {
                echo "${in}" | wc
                echo "$USER" > /dev/null
                echo "$(env)" >> /dev/null
            }
            output {
                String ans = stdout()
            }
        }
        """,
        """
        #
        task wc {
            # comment
            input {
                String in
            }
            String d = in + "_foo"
            command <<<
                echo "~{in}" | wc
                echo "$USER" > /dev/null
                echo "$(env)" >> /dev/null
                echo "${HOME}" >> /dev/null
            >>> #
            # comment
            output { #comment
                String ans = stdout()
            }
        }
        """]
        for task_str in variants:
            task = WDL.parse_tasks(task_str)[0]
            self.assertEqual(len(task.inputs), 1)
            self.assertEqual(str(task.inputs[0]), "String in")

            self.assertEqual(len(task.command.parts), 3)
            self.assertEqual(task.command.parts[1].expr._ident, "in")

            self.assertEqual(len(task.outputs), 1)
            self.assertEqual(str(task.outputs[0].type), "String")
            self.assertEqual(str(task.outputs[0].name), "ans")
            self.assertIsInstance(task.outputs[0].expr, WDL.Expr.Apply)

            task.typecheck()

            self.assertEqual(task.command.parts[1].eval(WDL.Env.Bindings(WDL.Env.Binding('in', WDL.Value.String("hello"))), WDL.StdLib.Base("1.0")).value, 'hello')

            self.assertFalse(task.command.parts[0].strip().startswith("{"))
            self.assertFalse(task.command.parts[0].strip().startswith("<<<"))
            self.assertFalse(task.command.parts[-1].strip().endswith("}"))
            self.assertFalse(task.command.parts[-1].strip().endswith(">>>"))

    def test_errors(self):
        with self.assertRaises(WDL.Error.InvalidType):
            WDL.parse_tasks("""
            task wc {
                input {
                    Int[Int] wrong
                }
                command {
                }
            }
            """)[0].typecheck()
        with self.assertRaises(WDL.Error.UnknownIdentifier, msg="Unknown identifier bogus"):
            WDL.parse_tasks("""
            task wc {
                input {
                    String in
                }
                command {
                    echo "~{bogus}" | wc
                }
            }
            """)[0].typecheck()
        with self.assertRaises(WDL.Error.UnknownIdentifier, msg="Unknown identifier bogus"):
            WDL.parse_tasks("""
            task wc {
                input {
                    String in
                }
                command {
                    echo "~{in}" | wc
                }
                output {
                    String ans = "${bogus}"
                }
            }
            """)[0].typecheck()
        with self.assertRaises(WDL.Error.MultipleValidationErrors):
            WDL.parse_tasks("""
            task wc {
                input {
                    String in
                    File out = stdout()
                    Array[File] files = glob("*.txt")
                }
                command {
                    echo "~{in}" | wc
                }
                output {
                }
            }
            """)[0].typecheck()

    def test_placeholders(self):
        task = WDL.parse_tasks("""
            task wc {
                input {
                    Boolean b
                }
                command <<<
                    echo "~{true='yes' false='no' b}"
                >>>
            }
            """)[0]
        task.typecheck()
        stdlib = WDL.StdLib.Base("1.0")
        self.assertEqual(task.command.parts[1].eval(WDL.Env.Bindings().bind('b', WDL.Value.Boolean(True)), stdlib).value, 'yes')
        self.assertEqual(task.command.parts[1].eval(WDL.Env.Bindings().bind('b', WDL.Value.Boolean(False)), stdlib).value, 'no')
        self.assertEqual(task.command.parts[1].eval(WDL.Env.Bindings().bind('b', WDL.Value.Null()), stdlib).value, '')

        task = WDL.parse_tasks("""
            task wc {
                input {
                    Boolean b
                    File? f
                }
                command {
                    echo "${if b then 'yes' else 'no'}"
                }
            }
            """)[0]
        task.typecheck()
        self.assertEqual(task.command.parts[1].eval(WDL.Env.Bindings().bind('b', WDL.Value.Boolean(True)), stdlib).value, 'yes')
        self.assertEqual(task.command.parts[1].eval(WDL.Env.Bindings().bind('b', WDL.Value.Boolean(False)), stdlib).value, 'no')
        with self.assertRaises(WDL.Error.NullValue):
            self.assertEqual(task.command.parts[1].eval(WDL.Env.Bindings().bind('b', WDL.Value.Null()), stdlib).value, '')

        with self.assertRaises(WDL.Error.StaticTypeMismatch):
            WDL.parse_tasks("""
                task wc {
                    input {
                        Int b
                    }
                    command {
                        echo "~{true='yes' false='no' b}"
                    }
                }
                """)[0].typecheck()

        with self.assertRaises(WDL.Error.StaticTypeMismatch):
            WDL.parse_tasks("""
                task wc {
                    command {
                        echo "~{true='yes' false='no' 42}"
                    }
                }
                """)[0].typecheck()

        with self.assertRaises(WDL.Error.StaticTypeMismatch):
            WDL.parse_tasks("""
                task wc {
                    input {
                        Boolean b
                    }
                    command {
                        echo "~{false='no' b}"
                    }
                }
                """)[0].typecheck()

        task = WDL.parse_tasks("""
            task wc {
                input {
                    Array[String] s
                }
                command <<<
                    echo "~{sep=', ' s} baz"
                >>>
            }
            """)[0]
        task.typecheck()
        foobar = WDL.Value.Array(WDL.Type.String(), [WDL.Value.String("foo"), WDL.Value.String("bar")])
        self.assertEqual(task.command.parts[1].eval(WDL.Env.Bindings().bind('s', foobar), stdlib).value, 'foo, bar')
        foobar = WDL.Value.Array(WDL.Type.String(), [])
        self.assertEqual(task.command.parts[1].eval(WDL.Env.Bindings().bind('s', foobar), stdlib).value, '')
        with self.assertRaises(WDL.Error.IncompatibleOperand):
            task = WDL.parse_tasks("""
            task wc {
                input {
                    Array[String] s
                }
                command <<<
                    echo "~{s} baz"
                >>>
            }
            """)[0].typecheck()
        with self.assertRaises(WDL.Error.StaticTypeMismatch):
            WDL.parse_tasks("""
            task wc {
                input {
                    String s
                }
                command <<<
                    echo "~{sep=', ' s} baz"
                >>>
            }
            """)[0].typecheck()

        task = WDL.parse_tasks("""
            task wc {
                input {
                    Boolean? b
                }
                command {
                    echo "${default='foo' b}"
                }
            }
            """)[0]
        task.typecheck()
        self.assertTrue(task.inputs[0].type.optional)
        self.assertEqual(task.command.parts[1].eval(WDL.Env.Bindings().bind('b', WDL.Value.Boolean(True)), stdlib).value, 'true')
        self.assertEqual(task.command.parts[1].eval(WDL.Env.Bindings().bind('b', WDL.Value.Boolean(False)), stdlib).value, 'false')
        self.assertEqual(task.command.parts[1].eval(WDL.Env.Bindings().bind('b', WDL.Value.Null()), stdlib).value, 'foo')

        task = WDL.parse_tasks("""
            task wc {
                input {
                    Boolean? b
                }
                output {
                    String ans = stdout()
                }
                command {
                    echo "${default='foo' b}"
                }
            }
            """)[0]
        task.typecheck()

        with self.assertRaises(WDL.Error.MultipleDefinitions):
            WDL.parse_tasks("""
                task wc {
                    input {
                        Boolean? b
                    }
                    output {
                        String ans = stdout()
                    }
                    command {
                        echo "${default='foo' b}"
                    }
                    output {
                        String ans2 = stdout()
                    }
                }
                """)[0]

    def test_meta(self):
        task = WDL.parse_tasks("""
        task wc {
            input {
                Boolean? b
                Array[Int]+ n
            }
            String dollar = "$"
            String lbrace = "{"
            String rbrace = "}"
            parameter_meta {
                b: { help: "it's a boolean" }
                n: 'x'
                IntKey: 999
            }
            command {
                echo "~{true='yes' false='no' b}"
            }
            runtime {
                memory: "1 GB"
                cpu: 42
                runtime_key: 123
                meta_key: 321
            }
        }
        """)[0]
        task.typecheck()
        self.assertIsInstance(task.parameter_meta['b']['help'], str)
        self.assertEqual(task.parameter_meta['b']['help'], "it's a boolean")
        self.assertIsInstance(task.parameter_meta['IntKey'], WDL.Expr.Int)
        self.assertIsInstance(task.runtime['cpu'], WDL.Expr.Int)
        self.assertEqual(task.runtime['cpu'].value, 42)
        self.assertEqual(task.runtime['runtime_key'].value, 123)
        self.assertEqual(task.runtime['meta_key'].value, 321)
        self.assertTrue(task.inputs[0].type.optional)
        self.assertFalse(task.inputs[1].type.optional)
        self.assertTrue(task.inputs[1].type.nonempty)

        task = WDL.parse_tasks(r"""
        task wc {
            input {
                Boolean? b
                Array[Int]+ n
            }
            String a_nonput_decl = "foo"
            meta {
                description: 'it\'s a task'
            }
            parameter_meta {
                b: { help: "it's a boolean" }
                n: 'x'
            }
            command {
                echo "~{true='yes' false='no' b}"
            }
            runtime {
                memory: "1 GB"
                cpu: 42
            }
        }
        """)[0]
        task.typecheck()
        self.assertIsInstance(task.meta['description'], str)
        self.assertEqual(task.meta['description'], "it's a task")

    def test_compare_md5sums(self):
        txt = """
task compare_md5sum {
	Array[String] labels
	Array[File] files
	Array[File] ref_files

	command <<<
		python <<CODE	
		from collections import OrderedDict
		import os
		import json
		import hashlib

		def md5sum(filename, blocksize=65536):
		    hash = hashlib.md5()
		    with open(filename, 'rb') as f:
		        for block in iter(lambda: f.read(blocksize), b""):
		            hash.update(block)
		    return hash.hexdigest()

		with open('${write_lines(labels)}','r') as fp:
			labels = fp.read().splitlines()
		with open('${write_lines(files)}','r') as fp:
			files = fp.read().splitlines()
		with open('${write_lines(ref_files)}','r') as fp:
			ref_files = fp.read().splitlines()

		result = OrderedDict()
		match = OrderedDict()
		match_overall = True

		result['tasks'] = []
		result['failed_task_labels'] = []
		result['succeeded_task_labels'] = []
		for i, label in enumerate(labels):
			f = files[i]
			ref_f = ref_files[i]
			md5 = md5sum(f)
			ref_md5 = md5sum(ref_f)
			# if text file, read in contents
			if f.endswith('.qc') or f.endswith('.txt') or \
				f.endswith('.log') or f.endswith('.out'):
				with open(f,'r') as fp:
					contents = fp.read()
				with open(ref_f,'r') as fp:
					ref_contents = fp.read()
			else:
				contents = ''
				ref_contents = ''
			matched = md5==ref_md5
			result['tasks'].append(OrderedDict([
				('label', label),
				('match', matched),
				('md5sum', md5),
				('ref_md5sum', ref_md5),
				('basename', os.path.basename(f)),
				('ref_basename', os.path.basename(ref_f)),
				('contents', contents),
				('ref_contents', ref_contents),
				]))
			match[label] = matched
			match_overall &= matched
			if matched:
				result['succeeded_task_labels'].append(label)
			else:
				result['failed_task_labels'].append(label)		
		result['match_overall'] = match_overall

		with open('result.json','w') as fp:
			fp.write(json.dumps(result, indent=4))
		match_tmp = []
		for key in match:
			val = match[key]
			match_tmp.append('{}\t{}'.format(key, val))
		with open('match.tsv','w') as fp:
			fp.writelines('\n'.join(match_tmp))
		with open('match_overall.txt','w') as fp:
			fp.write(str(match_overall))
		CODE
	>>>
	output {
		Map[String,String] match = read_map('match.tsv') # key:label, val:match
		Boolean match_overall = read_boolean('match_overall.txt')
		File json = glob('result.json')[0] # details (json file)
		String json_str = read_string('result.json') # details (string)
	}
	runtime {
		cpu : 1
		memory : "4000 MB"
		time : 1
		disks : "local-disk 50 HDD"		
	}
}
"""
        task = WDL.parse_tasks(txt, version="draft-2")[0]
        task.typecheck()
        self.assertEqual(len(task.command.parts), 7)
        self.assertFalse(task.command.parts[0].strip().startswith("<<<"))
        self.assertFalse(task.command.parts[-1].strip().endswith(">>>"))


class TestTypes(unittest.TestCase):
    def test_parser(self):
        def check(t,v):
            doc_txt = r"""
            workflow contrived {{
                {t} x
                {t} y
                output {{
                    {t} o_y = y
                }}
            }}
            """.format(t=t)
            doc = WDL.parse_document(doc_txt, v)
            self.assertEqual(str(doc.workflow.body[0].type), t)
            self.assertEqual(doc.workflow.body[0].type.optional, t.endswith("?"))
            self.assertEqual(str(list(doc.workflow.effective_outputs)[0].value), t)
        for t in ["Int", "Int?",
                  "Array[Int]", "Array[Int]?", "Array[Int]+?", "Array[Int?]+?",
                  "Map[String,Int]", "Map[String,Array[Float?]+]?",
                  "Pair[Int,Float?]"]:
            check(t, "draft-2")
            check(t, "1.0")

    def test_invalid(self):
        def check(t,v):
            doc_txt = r"""
            workflow contrived {{
                {t} x
            }}
            """.format(t=t)
            WDL.parse_document(doc_txt, v)
        for t in ["Int+", "Array?+", "Array[Int,Float]", "Map[String]", "Map[Int,Int,Int]",
                  "Pair[Int]", "Pair[Float,Float,Float]", "Array[Pair[Int]?]+",
                  "Bogus", "bogus?"]:
            with self.assertRaises((WDL.Error.InvalidType, WDL.Error.SyntaxError)):
                check(t,"draft-2")
                check(t,"1.0")

    def test_unify(self):
        doc = WDL.parse_document("""
        workflow unify {
            String s
            File? f2
            Array[Int] a1 = [1]
            Array[Int?]? a2 = []
            Array[Pair[String,String]] ap = [(0,1),(2,3)]
            Array[Map[String,String]] am = [{ "a": 0, "b": 1 }, { "a": "x", "b": "y" }, { 1: 2, 3: 4 }]

            output {
                Array[File?] a = [s, f2]
                Array[Array[Int?]?] a3 = [a1, a2]
                Map[String, File?] m = { "foo": s, "bar": f2 }
                Map[Float, File?] m2 = { 1: s, 2.0: f2 }
            }
        }
        """)
        doc.typecheck()
        for outp in doc.workflow.outputs:
            self.assertIsNone(outp.expr.literal)

        with self.assertRaises(WDL.Error.ValidationError):
            doc = WDL.parse_document("""
            workflow unify {
                Array[Pair[String,String]] bogus = [("a","b"), ("c",("d","e"))]
            }
            """)
            doc.typecheck()

    def test_map_io(self):
        with self.assertRaisesRegex(WDL.Error.ValidationError, "keys cannot"):
            WDL.parse_document("""
            workflow w {
                input {
                    Map[Pair[Int,Int],String] m
                }
            }
            """).typecheck()

        with self.assertRaisesRegex(WDL.Error.ValidationError, "keys cannot"):
            WDL.parse_document("""
            task t {
                command {}
                output {
                    Map[Pair[Int,Int],String] m = read_json("bogus")
                }
            }
            """).typecheck()

class TestDoc(unittest.TestCase):
    def test_count_foo(self):
        doc = r"""#foo
workflow count_lines_matching {
    call grep
    call count_lines {
        input:
            in = grep.out  	# bar
    }
}
task grep {
    File in
    String pattern

    command {
        grep ${pattern} ${in} > ans
    }

    output {
        File out = "ans"
    }
}
    #baz   
task count_lines {
    File in

    command {
        wc -l ${in}
    }

    output {
        Int out = read_int(stdout())
    }
}
   #bas 
"""
        doc = WDL.parse_document(doc)
        self.assertIsInstance(doc.workflow, WDL.Tree.Workflow)
        self.assertEqual(len(doc.workflow.body), 2)
        self.assertEqual(len(doc.tasks), 2)
        doc.typecheck()
        self.assertEqual(doc.source_comments[0].text, "#foo")
        self.assertEqual(doc.source_comments[5].text, "  	# bar")
        self.assertEqual(doc.source_comments[20].text, "    #baz   ")
        self.assertEqual(doc.source_comments[32].text, "   #bas ")

    def test_bam_chrom_counter(self):
        doc = r"""
        workflow bam_chrom_counter {
            File bam

            call slice_bam {
                input : bam = bam
            }
            scatter (slice in slice_bam.slices) {
                call count_bam {
                    input: bam = slice
                }
            }
            output {
                File bai = slice_bam.bai
                Array[Int] count = count_bam.count
            }
        }

        task slice_bam {
            File bam
            Int num_chrom = 22
            command <<<
                set -ex
                samtools index ${bam}
                mkdir slices/
                for i in `seq ${num_chrom}`; do
                    samtools view -b ${bam} -o slices/$i.bam $i
                done
            >>>
            runtime {
                docker: "quay.io/ucsc_cgl/samtools"
            }
            output {
                File bai = "${bam}.bai"
                Array[File] slices = glob("slices/*.bam")
            }
        }

        task count_bam {
            File bam
            command {
                samtools view -c ${bam}
            }
            runtime {
                docker: "quay.io/ucsc_cgl/samtools"
            }
            output {
                Int count = read_int(stdout())
            }
        }
        """
        doc = WDL.parse_document(doc, version="draft-2")
        self.assertIsInstance(doc.workflow, WDL.Tree.Workflow)
        self.assertEqual(len(doc.workflow.body), 3)
        self.assertIsInstance(doc.workflow.body[2], WDL.Tree.Scatter)
        self.assertEqual(len(doc.workflow.body[2].body), 1)
        self.assertEqual(len(doc.tasks), 2)
        self.assertEqual(doc.tasks[0].name, "slice_bam")
        self.assertEqual(len(doc.tasks[0].command.parts), 7)
        self.assertFalse(doc.tasks[0].command.parts[0].strip().startswith("<<<"))
        self.assertFalse(doc.tasks[0].command.parts[-1].strip().endswith(">>>"))
        self.assertFalse(doc.tasks[1].command.parts[0].strip().startswith("{"))
        self.assertFalse(doc.tasks[1].command.parts[-1].strip().endswith("}"))
        doc.typecheck()

    def test_nested_scatter(self):
        doc = r"""
        import "x.wdl"
        task sum {
            Int x
            Int y
            command <<<
                echo $(( ~{x} + ~{y} ))
            >>>
            output {
                Int z = read_int(stdout())
            }
        }
        import "y.wdl" as z
        workflow contrived {
            Array[Int] xs = [1, 2, 3]
            Array[Int] ys = [4, 5, 6]
            scatter (x in xs) {
                scatter (y in ys) {
                    call sum { input:
                        x = x,
                        y = y
                    }
                }
            }
            output {
                Array[Array[Int]] z = sum.z
            }
        }
        """
        doc = WDL.parse_document(doc)
        self.assertIsInstance(doc.workflow, WDL.Tree.Workflow)
        self.assertIsInstance(doc.workflow.body[2], WDL.Tree.Scatter)
        self.assertIsInstance(doc.workflow.body[2].body[0], WDL.Tree.Scatter)
        self.assertEqual(len(doc.tasks), 1)
        doc.typecheck()
        self.assertEqual(len(doc.imports), 2)
        self.assertEqual(doc.imports[0].uri, "x.wdl")
        self.assertEqual(doc.imports[0].namespace, "x")
        self.assertEqual(doc.imports[1].uri, "y.wdl")
        self.assertEqual(doc.imports[1].namespace, "z")

    def test_scatter_conditional(self):
        doc = r"""
        task sum {
            Int x
            Int y
            command <<<
                echo $(( ~{x} + ~{y} ))
            >>>
            output {
                Int z = read_int(stdout())
            }
            meta {
                foo: "bar"
            }
        }
        workflow contrived {
            Array[Int] xs = [1, 2, 3]
            Array[Int] ys = [4, 5, 6]
            scatter (x in xs) {
                Int x2_ = x*x
                scatter (y in ys) {
                    if (x + y < 5) {
                        Int xy_ = x * y
                        call sum { input:
                            x = x,
                            y = y
                        }
                    }
                }
            }
            output {
                Array[Array[Int?]] z = sum.z
                Array[Array[Int?]] xy = xy_
                Array[Int] x2 = x2_
            }
        }
        """
        doc = WDL.parse_document(doc)
        doc.typecheck()

        b = [False]
        def check_scatter_depth(node, d=0):
            self.assertEqual(node.scatter_depth, d)
            if d>0:
                b[0] = True
            for ch in node.children:
                if isinstance(ch, WDL.Tree.WorkflowNode):
                    check_scatter_depth(ch, d=(d+1 if isinstance(node, WDL.Scatter) and not isinstance(ch, WDL.Tree.Gather) else d))
        for node in doc.workflow.body:
            check_scatter_depth(node)
        assert b[0]

    def test_errors(self):
        doc = r"""
        version 1.0
        task sum {
            Int x
            Int y
            command <<<
                echo $(( ~{x} + ~{y} ))
            >>>
            output {
                Int z = read_int(stdout())
            }
        }
        workflow contrived {
            Int not_array
            scatter (x in not_array) {
            }
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.NotAnArray):
            doc.typecheck()

        doc = r"""
        version 1.0
        task sum {
            Int x
            Int y
            command <<<
                echo $(( ~{x} + ~{y} ))
            >>>
            output {
                Int z = read_int(stdout())
            }
        }
        workflow contrived {
            call sum { input:
                x = 1,
                z = 0
            }
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.NoSuchInput):
            doc.typecheck()

        doc = r"""
        version 1.0
        task sum {
            Int x
            Int y
            command <<<
                echo $(( ~{x} + ~{y} ))
            >>>
            output {
                Int z = read_int(stdout())
            }
        }
        workflow contrived {
            call sum { input:
                x = 1,
                y = 2
            }
            output {
                Int z = sum.bogus
            }
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaisesRegex(WDL.Error.UnknownIdentifier, " in namespace "):
            doc.typecheck()

        doc = r"""
        workflow contrived {
            call bogus { input:
                x = 1,
                y = 2
            }
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.NoSuchTask):
            doc.typecheck()

        doc = r"""
        version 1.0
        task sum {
            Int x
            Int y
            command <<<
                echo $(( ~{x} + ~{y} ))
            >>>
            output {
                Int z = read_int(stdout())
            }
        }
        workflow contrived {
            if ([1]) {
                call sum { input:
                    x = 1, y = 2
                }
            }
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.StaticTypeMismatch):
            doc.typecheck()

        doc = r"""
        version 1.0
        task sum {
            Int x
            Int y
            command <<<
                echo $(( ~{x} + ~{y} ))
            >>>
            output {
                Int z = read_int(stdout())
            }
        }
        task sum {
            command {}
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.MultipleDefinitions):
            doc.typecheck()

        doc = r"""
        version 1.0
        task sum {
            Int x
            Int y
            command <<<
                echo $(( ~{x} + ~{y} ))
            >>>
            output {
                Int z = read_int(stdout())
            }
        }
        workflow sum {
            call sum
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.MultipleDefinitions):
            doc.typecheck()

        doc = r"""
        version 1.0
        task sum {
            Int x
            Int y
            File y
            command <<<
                echo $(( ~{x} + ~{y} ))
            >>>
            output {
                Int z = read_int(stdout())
            }
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.MultipleDefinitions):
            doc.typecheck()

        doc = r"""
        workflow contrived {
            Int x
            if (true) {
                Int x = 1
            }
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.MultipleDefinitions):
            doc.typecheck()

        doc = r"""
        workflow contrived {
            Int x
            output {
                Int x = 1
            }
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.MultipleDefinitions):
            doc.typecheck()

        doc = r"""
        import "x.wdl"
        import "x.y.wdl"
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.MultipleDefinitions):
            doc.typecheck()

        doc = r"""
        import "x.wdl"
        import "y.wdl" as x
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.MultipleDefinitions):
            doc.typecheck()

        doc = r"""
        import "https://www.google.com/task.wdl"
        """
        with self.assertRaises(WDL.Error.SyntaxError):
            WDL.parse_document(doc)

        doc = r"""
        import "0"
        """
        with self.assertRaises(WDL.Error.SyntaxError):
            WDL.parse_document(doc)

        doc = r"""
        workflow wf {
            scatter (x in []) {
            }
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.IndeterminateType):
            doc.typecheck()

    def test_task_forward_reference(self):
        doc = r"""
        version 1.0
        task sum {
            input {
                Int x = y
            }
            Int y = 42
            command <<<
                echo $(( ~{x} + ~{y} ))
            >>>
            output {
                Int z = read_int(stdout())
            }
        }
        """
        doc = WDL.parse_document(doc)
        doc.typecheck()
        self.assertEqual(len(doc.tasks[0].command.parts), 5)

        # TODO: test circular reference

    def test_draft2_workflow_outputs(self):
        doc = r"""
        task sum {
            Int x
            Int y
            command <<<
                echo $(( ~{x} + ~{y} ))
            >>>
            output {
                Int z = read_int(stdout())
            }
            meta {
                foo: "bar"
            }
        }
        workflow contrived {
            Array[Int] xs = [1, 2, 3]
            Array[Int] ys = [4, 5, 6]
            scatter (x in xs) {
                Int x2_ = x*x
                scatter (y in ys) {
                    if (x + y < 5) {
                        Int xy_ = x * y
                        call sum { input:
                            x = x,
                            y = y
                        }
                    }
                }
            }
            output {
                sum.z
            }
        }
        """
        doc = WDL.parse_document(doc)
        doc.typecheck()
        self.assertEqual(str(doc.workflow.outputs[0].type), "Array[Array[Int?]]")

        doc = WDL.parse_document("""
            workflow bogus {
                output {
                    z
                }
            }
        """)
        with self.assertRaises(WDL.Error.UnknownIdentifier):
            doc.typecheck()

        doc = WDL.parse_document("""
            workflow bogus {
                output {
                    add.z
                }
            }
        """)
        with self.assertRaises(WDL.Error.UnknownIdentifier):
            doc.typecheck()

        doc = WDL.parse_document("""
            workflow bogus {
                output {
                    nonex.*
                }
            }
        """)
        with self.assertRaises(WDL.Error.NoSuchTask):
            doc.typecheck()

        doc = WDL.parse_document("""
            task sum {
                Int x
                Int y
                command <<<
                    echo $(( ~{x} + ~{y} ))
                >>>
                output {
                    Int z = read_int(stdout())
                }
                meta {
                    foo: "bar"
                }
            }
            workflow bogus {
                call sum as adder
                output {
                    Int z = 4
                    adder.z
                }
            }
        """)
        doc.typecheck()

        doc = WDL.parse_document("""
            task sum {
                Int x
                Int y
                command <<<
                    echo $(( ~{x} + ~{y} ))
                >>>
                output {
                    Int z = read_int(stdout())
                }
                meta {
                    foo: "bar"
                }
            }
            workflow bogus {
                call sum
                call sum as adder
                output {
                    sum.z
                    adder.z
                }
            }
        """)
        doc.typecheck()

        doc = WDL.parse_document("""
            task sum {
                Int x
                Int y
                command <<<
                    echo $(( ~{x} + ~{y} ))
                >>>
                output {
                    Int z = read_int(stdout())
                }
                meta {
                    foo: "bar"
                }
            }
            workflow bogus {
                call sum as adder
                output {
                    adder.z
                    adder.z
                }
            }
        """)
        with self.assertRaises(WDL.Error.MultipleDefinitions):
            doc.typecheck()

        try:
            doc = WDL.parse_document("""
                version 1.0
                task sum {
                    Int x
                    Int y
                    command <<<
                        echo $(( ~{x} + ~{y} ))
                    >>>
                    output {
                        Int z = read_int(stdout())
                    }
                    meta {
                        foo: "bar"
                    }
                }
                workflow bogus {
                    call sum as adder
                    output {
                        adder.z
                    }
                }
            """)
            assert False
        except WDL.Error.SyntaxError as err:
            self.assertEqual(err.pos.line, 19)
            self.assertEqual(err.pos.column, 30)

        doc = WDL.parse_document("""
            task sum {
                Int x
                Int y
                command <<<
                    echo $(( ~{x} + ~{y} ))
                >>>
                output {
                    Int z = read_int(stdout())
                    Int w = 4
                }
                meta {
                    foo: "bar"
                }
            }
            workflow bogus {
                call sum
                call sum as adder
                output {
                    sum.*
                    adder.*
                }
            }
        """)
        doc.typecheck()
        self.assertEqual(set(decl.name for decl in doc.workflow.outputs),
                         set(["adder.w", "adder.z", "sum.w", "sum.z"]))

        doc = WDL.parse_document("""
            task sum {
                Int x
                Int y
                command <<<
                    echo $(( ~{x} + ~{y} ))
                >>>
                output {
                    Int z = read_int(stdout())
                    Int w = 4
                }
                meta {
                    foo: "bar"
                }
            }
            workflow bogus {
                call sum
                call sum as adder
                Int k = 4
                Int j = 5
                output {
                    sum.*
                    adder.*
                    k
                    Int j2 = j
                }
            }
        """)
        doc.typecheck()

    def test_workflow_inputs(self):
        doc = r"""
        version 1.0
        task sum {
            Int x
            Int y
            command <<<
                echo $(( ~{x} + ~{y} ))
            >>>
            output {
                Int z = read_int(stdout())
            }
            meta {
                foo: "bar"
            }
        }
        workflow contrived {
            input {
                Int x
                Int y
            }
            call sum { input:
                x = x,
                y = y
            }
            output {
                Int z = sum.z
            }
        }
        """
        doc = WDL.parse_document(doc)
        doc.typecheck()
        self.assertEqual(set(decl.name for decl in doc.workflow.inputs), set(["x", "y"]))

        doc = r"""
        version 1.0
        task sum {
            input {
                Int x = y
            }
            Int y
            command <<<
                echo $(( ~{x} + ~{y} ))
            >>>
            output {
                Int z = read_int(stdout())
            }
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.StrayInputDeclaration):
            doc.typecheck()

        doc = r"""
        version 1.0
        task sum {
            input {
                Int x = select_first([y,42])
            }
            Int? y
            command <<<
                echo $(( ~{x} + ~{y} ))
            >>>
            output {
                Int z = read_int(stdout())
            }
        }
        """
        doc = WDL.parse_document(doc)
        doc.typecheck()
        self.assertIn("x", doc.tasks[0].available_inputs)
        self.assertNotIn("x", doc.tasks[0].required_inputs)
        self.assertNotIn("y", doc.tasks[0].available_inputs)

        doc = r"""
        workflow wf {
            input {
                Int x = y
            }
            Int y
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.StrayInputDeclaration):
            doc.typecheck()

        doc = r"""
        workflow wf {
            input {
                Int x = select_first([y, 42])
            }
            Int? y
        }
        """
        doc = WDL.parse_document(doc)
        doc.typecheck()
        self.assertIn("x", doc.workflow.available_inputs)
        self.assertNotIn("y", doc.workflow.available_inputs)

        doc = r"""
        workflow wf {
            input {
                Int x = y[0]
            }
            scatter (i in [1,2,3]) {
                Int y
            }
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.StrayInputDeclaration):
            doc.typecheck()

        doc = r"""
        workflow wf {
            scatter (i in [1,2,3]) {
                Int y
            }
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.StrayInputDeclaration):
            doc.typecheck()

        doc = r"""
        workflow wf {
            input {
                Int x = select_first([y, 42])
            }
            Int? y
        }
        """
        doc = WDL.parse_document(doc)
        doc.typecheck()
        self.assertIn("x", doc.workflow.available_inputs)
        self.assertNotIn("y", doc.workflow.available_inputs)

        doc = r"""
        workflow wf {
            Int x
            scatter (i in [1,2,3]) {
                Int? y
            }
        }
        """
        doc = WDL.parse_document(doc)
        doc.typecheck()
        self.assertIn("x", doc.workflow.available_inputs)
        self.assertNotIn("y", doc.workflow.available_inputs)

        doc = r"""
        version 1.0
        task sum {
            input {
                Int x
                Int y
            }
            command <<<
                echo $(( ~{x} + ~{y} ))
            >>>
            output {
                Int z = read_int(stdout())
            }
        }
        workflow wf {
            input {
                Int x = y
                Int y
                Int z = sum.z
            }
            call sum { input:
                x = x,
                y = y
            }
            output {
                Int z2 = z+1
                Int w = x+y
                Array[Int] outs = [z,w]
            }
        }
        """
        doc = WDL.parse_document(doc)
        doc.typecheck()

    def test_multi_errors(self):
        doc = r"""
        version 1.0
        task bogus1 {
            Int x = false
            command {
            }
        }
        task bogus2 {
            input {
                Int x
                Int y
            }
            command <<<
                echo $(( ~{x} + ~{y} ))
            >>>
            output {
                Int z = read_int(stdout())
                Int w = 17 + false
            }
        }
        """
        doc = WDL.parse_document(doc)
        try:
            doc.typecheck()
            assert False
        except WDL.Error.MultipleValidationErrors as multi:
            self.assertEqual(len(multi.exceptions), 2)

        doc = r"""
        version 1.0
        workflow bogus {
            Int x = false
            call sum { input:
                not_there = 21
            }
            output {
                File wrong = sum.z
                Int bogus = sum.not_there
            }
        }
        task sum {
            input {
                Int x
                Int y
            }
            command <<<
                echo $(( ~{x} + ~{y} ))
            >>>
            output {
                Int z = read_int(stdout())
            }
        }
        """
        doc = WDL.parse_document(doc)
        try:
            doc.typecheck()
            assert False
        except WDL.Error.MultipleValidationErrors as multi:
            self.assertEqual(len(multi.exceptions), 4)

        doc = r"""
        version 1.0
        workflow bogus {
            call not_there_1
            call not_there_2
        }
        """
        doc = WDL.parse_document(doc)
        try:
            doc.typecheck()
            assert False
        except WDL.Error.MultipleValidationErrors as multi:
            self.assertEqual(len(multi.exceptions), 2)

        doc = r"""
        version 1.0
        workflow bogus {
            File f
            Array[String] a = [f+1,f+2]
        }
        """
        doc = WDL.parse_document(doc)
        try:
            doc.typecheck()
            assert False
        except WDL.Error.MultipleValidationErrors as multi:
            self.assertEqual(len(multi.exceptions), 2)

    def test_issue135_workflow_available_inputs(self):
        # Workflow.available_inputs should not include declarations in the
        # output section
        doc = r"""
        workflow a {
            File in
            output {
                File out = in
            }
        }
        """
        doc = WDL.parse_document(doc)
        doc.typecheck()
        self.assertEqual(len(doc.workflow.available_inputs), 1)
        self.assertEqual(list(doc.workflow.available_inputs)[0].name, "in")

    def test_issue173_workflow_section_order(self):
        doc = r"""
        version 1.0
        workflow a {
            output {
                String d = c
            }
            String c = b
            input {
                String b
            }
            parameter_meta {
                b: "Test"
            }
            meta {
                test: "This is a test"
            }
        }
        """
        WDL.parse_document(doc).typecheck()

    def test_issue233_version_development(self):
        # bug where "version development" resulted in draft-2 grammar
        doc = r"""
        version development

        # struct won't parse under draft-2
        struct Car {
            String make
            String model
        }
        """
        WDL.parse_document(doc).typecheck()

class TestCycleDetection(unittest.TestCase):
    def test_task(self):
        doc = r"""
        version 1.0
        task cyclic {
            input {
                Int i
            }
            Int x = i + y
            Int y = i - x

            command{}
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.CircularDependencies):
            doc.typecheck()

        doc = r"""
        version 1.0
        task cyclic {
            input {
                Int i
            }
            Int x = i + y
            Int y = i - z
            Int z = i * x

            command{}
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.CircularDependencies):
            doc.typecheck()

        doc = r"""
        version 1.0
        task cyclic {
            input {
                Int i = i
            }

            command{}
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.CircularDependencies):
            doc.typecheck()

    def test_workflow(self):
        add = r"""
        task add {
            input {
                Int lhs
                Int rhs
            }

            command{}

            output {
                Int z = lhs + rhs
            }
        }
        """
        doc = r"""
        version 1.0
        workflow cyclic {
            input {
                Int n = add.z
            }
            call add { input: lhs = 0, rhs = n }
        }
        """ + add
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.CircularDependencies):
            doc.typecheck()

        doc = r"""
        version 1.0
        workflow cyclic {
            input {
                Int n = add.z
            }
            call add { input: lhs = 1, rhs = add2.z }
            call add as add2 { input: lhs = n, rhs = 0 }
        }
        """ + add
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.CircularDependencies):
            doc.typecheck()

        doc = r"""
        version 1.0
        workflow cyclic {
            input {
                Boolean b
            }
            scatter (i in [1, 2, 3]) {
                call add { input: lhs = i, rhs = select_first([add2.z,0]) }
            }
            if (b) {
                call add as add2 { input: lhs = add.z[0], rhs = 0 }
            }
        }
        """ + add
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.CircularDependencies):
            doc.typecheck()

        doc = r"""
        version 1.0
        workflow cyclic {
            if (select_first([c])) {
                Boolean b = true
            }
            if (select_first([b])) {
                Boolean c = true
            }
        }
        """ + add
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.CircularDependencies):
            doc.typecheck()

        doc = r"""
        version 1.0
        workflow cyclic {
            scatter (i in arr) {
                Int j = i
            }
            Array[Int] arr = j
        }
        """ + add
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.CircularDependencies):
            doc.typecheck()

        doc = r"""
        version 1.0
        workflow cyclic {
            call add { input: lhs = 0, rhs = add.z }
        }
        """ + add
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.CircularDependencies):
            doc.typecheck()

class TestStruct(unittest.TestCase):
    def test_parser(self):
        doc = r"""
        version 1.0

        struct Person {
            String name
            Int age
        }

        struct Name {
            Array[File]+ myFiles
            Boolean? myBoolean
        }
        """
        doc = WDL.parse_document(doc)
        doc.typecheck()
        self.assertEqual(str(doc.struct_typedefs.resolve("Person").members["age"]), "Int")
        self.assertEqual(str(doc.struct_typedefs.resolve("Name").members["myFiles"]), "Array[File]+")

        doc = r"""
        version 1.0

        struct Person {
            String a
            Int a
        }
        """
        with self.assertRaises(WDL.Error.MultipleDefinitions):
            doc = WDL.parse_document(doc)

        doc = r"""
        version 1.0

        struct Person {
            String a
        }

        struct Person {
            Int b
        }
        """
        with self.assertRaises(WDL.Error.MultipleDefinitions):
            doc = WDL.parse_document(doc)

    def test_decl(self):
        doc = r"""
        version 1.0

        workflow UsePerson {
            Person p
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.InvalidType):
            doc.typecheck()

        doc = r"""
        version 1.0

        workflow UsePerson {
            Person p
            Person p2 = p
        }

        struct Person {
            String name
            Int age
        }
        """
        doc = WDL.parse_document(doc)
        doc.typecheck()

        doc = r"""
        version 1.0

        workflow UsePerson {
            Person p
            Person p2 = 0
        }

        struct Person {
            String name
            Int age
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.StaticTypeMismatch):
            doc.typecheck()

        doc = r"""
        version 1.0

        workflow UsePerson {
            Person p
            Int k = p
        }

        struct Person {
            String name
            Int age
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.StaticTypeMismatch):
            doc.typecheck()

        doc = r"""
        version 1.0

        workflow UsePerson {
            Person p
            Car c
            Person p2 = c
        }

        struct Person {
            String name
            Int age
        }

        struct Car {
            String make
            String model
            Int year
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaisesRegex(WDL.Error.StaticTypeMismatch, "Expected Person instead of Car"):
            doc.typecheck()

        doc = r"""
        version 1.0

        workflow UsePerson {
            Person p
            Car c
        }

        struct Person {
            String name
            Int age
        }

        struct Car {
            String make
            String model
            Int year
            Person driver
        }
        """
        doc = WDL.parse_document(doc)
        doc.typecheck()

        doc = r"""
        version 1.0

        workflow UsePerson {
            Person p
            Car c
        }

        struct Person {
            String name
            Int age
            Car vehicle
        }

        struct Car {
            String make
            String model
            Int year
            Person driver
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.CircularDependencies):
            doc.typecheck()

        doc = r"""
        version 1.0

        struct Self {
            Int k
            Self me
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.CircularDependencies):
            doc.typecheck()

        doc = r"""
        version 1.0

        workflow UsePerson {
            Car c
            Int year = c.year
            Int age = c.driver.age
            Int month = c.driver.birthday.left
        }

        struct Person {
            String name
            Int age
            Pair[Int,Int] birthday
        }

        struct Car {
            String make
            String model
            Int year
            Person driver
        }
        """
        doc = WDL.parse_document(doc)
        doc.typecheck()

        doc = r"""
        version 1.0

        workflow UsePerson {
            Person? p
            Int age = p.age
        }

        struct Person {
            String name
            Int age
            Pair[Int,Int] birthday
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.StaticTypeMismatch):
            doc.typecheck()

        doc = r"""
        version 1.0

        workflow UsePerson {
            Person? p
            Int? age = p.age
        }

        struct Person {
            String name
            Int age
            Pair[Int,Int] birthday
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.StaticTypeMismatch):
            doc.typecheck()

        doc = r"""
        version 1.0

        workflow UsePerson {
            Person? p
            Int month = p.birthday.left
        }

        struct Person {
            String name
            Int age
            Pair[Int,Int] birthday
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.StaticTypeMismatch):
            doc.typecheck()

        doc = r"""
        version 1.0

        workflow UsePerson {
            Person p
            Int month = p.bogus
        }

        struct Person {
            String name
            Int age
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.NoSuchMember):
            doc.typecheck()

        doc = r"""
        version 1.0

        workflow UsePerson {
            Person p
            Int month = p.left
        }

        struct Person {
            String name
            Int age
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.NoSuchMember):
            doc.typecheck()

        doc = r"""
        version 1.0

        workflow UsePerson {
            Array[Person]+ ps

            output {
                Int p0age = ps[0].age
            }
        }

        struct Person {
            String name
            Int age
        }
        """
        doc = WDL.parse_document(doc)
        doc.typecheck()

        doc = r"""
        version 1.0

        workflow UsePerson {
            Array[Person]+ ps

            output {
                Int q0age = q[0].age
            }
        }

        struct Person {
            String name
            Int age
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.UnknownIdentifier):
            doc.typecheck()

        doc = r"""
        version 1.0

        workflow UsePerson {
            Person[Int,Float] bogus
        }

        struct Person {
            String name
            Int age
        }
        """
        with self.assertRaises(WDL.Error.InvalidType):
            WDL.parse_document(doc)

        # Test struct members with compound types including other structs
        # (issue #127)
        doc = r"""
        version 1.0

        workflow UsePerson {
            Person p
        }

        struct Person {
            String name
            Array[Pair[Car,Int]] cars
        }

        struct Car {
            String make
            String model
        }
        """
        doc = WDL.parse_document(doc)
        doc.typecheck()

        doc = r"""
        version 1.0

        workflow UsePerson {
            Person p
        }

        struct Person {
            String name
            Array[Person] circular
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.CircularDependencies):
            doc.typecheck()

        doc = r"""
        version 1.0

        workflow UsePerson {
            Person p
        }

        struct Person {
            String name
            Array[Car] cars
        }

        struct Car {
            String make
            String model
            Person owner
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.CircularDependencies):
            doc.typecheck()

    def test_import(self):
        doc = r"""
        version 1.0
        import "../test_corpi/gatk-workflows/five-dollar-genome-analysis-pipeline/structs/GermlineStructs.wdl"
        """
        docfn = tempfile.mktemp(".wdl")
        with open(docfn,"w") as outfile:
            outfile.write(doc)
        doc = WDL.load(docfn, path=[os.path.dirname(__file__)])

        doc = r"""
        version 1.0
        import "../test_corpi/gatk-workflows/five-dollar-genome-analysis-pipeline/structs/GermlineStructs.wdl"

        struct SampleAndUnmappedBams {
            Int bogus
        }
        """
        with open(docfn,"w") as outfile:
            outfile.write(doc)
        with self.assertRaises(WDL.Error.MultipleDefinitions):
            doc = WDL.load(docfn, path=[os.path.dirname(__file__)])

        doc = r"""
        version 1.0
        import "../test_corpi/gatk-workflows/five-dollar-genome-analysis-pipeline/structs/GermlineStructs.wdl" alias SampleAndUnmappedBams as SAUB

        struct SampleAndUnmappedBams {
            Int bogus
        }
        """
        with open(docfn,"w") as outfile:
            outfile.write(doc)
        doc = WDL.load(docfn, path=[os.path.dirname(__file__)])

        doc = r"""
        version 1.0
        import "../test_corpi/gatk-workflows/five-dollar-genome-analysis-pipeline/structs/GermlineStructs.wdl" alias PapiSettings as SampleAndUnmappedBams
        """
        with open(docfn,"w") as outfile:
            outfile.write(doc)
        with self.assertRaises(WDL.Error.MultipleDefinitions):
            doc = WDL.load(docfn, path=[os.path.dirname(__file__)])

        doc = r"""
        version 1.0
        import "../test_corpi/gatk-workflows/five-dollar-genome-analysis-pipeline/structs/GermlineStructs.wdl" alias PapiSettings as Person

        struct Person {
            String name
        }
        """
        with open(docfn,"w") as outfile:
            outfile.write(doc)
        with self.assertRaises(WDL.Error.MultipleDefinitions):
            doc = WDL.load(docfn, path=[os.path.dirname(__file__)])

        doc = r"""
        version 1.0
        import "../test_corpi/gatk-workflows/five-dollar-genome-analysis-pipeline/structs/GermlineStructs.wdl" alias PapiSettings as PapiSettings
        """
        with open(docfn,"w") as outfile:
            outfile.write(doc)
        with self.assertRaises(WDL.Error.MultipleDefinitions):
            doc = WDL.load(docfn, path=[os.path.dirname(__file__)])

        doc = r"""
        version 1.0
        import "../test_corpi/gatk-workflows/five-dollar-genome-analysis-pipeline/structs/GermlineStructs.wdl" alias Bogus as AlsoBogus
        """
        with open(docfn,"w") as outfile:
            outfile.write(doc)
        with self.assertRaises(WDL.Error.NoSuchMember):
            doc = WDL.load(docfn, path=[os.path.dirname(__file__)])

        os.unlink(docfn)

    def test_object_literal(self):
        doc = r"""
        version 1.0

        workflow wf {
            Person alyssa = object { name: "alyssa", "age": 42, 'birthday': (4,20) }
            call t { input: p = alyssa }
        }

        task t {
            input {
                Person p
                Car c
            }

            command <<<
                echo "Hello, ~{p.name}!"
            >>>

            output {
                Person p2 = object { name: stdout(), age: 99, birthday: (1,23) }
            }
        }

        struct Person {
            String name
            Int age
            Pair[Int,Int] birthday
        }

        struct Car {
            String make
            String model
            Int year
            Person driver
        }
        """
        doc = WDL.parse_document(doc)
        doc.typecheck()
        self.assertIsInstance(doc.workflow.body[0].expr.literal, WDL.Value.Struct)
        self.assertIsNone(doc.tasks[0].outputs[0].expr.literal)

        # quickly check task pickle-ability with struct types
        pickled_task = pickle.dumps(doc.tasks[0])
        unpickled_task = pickle.loads(pickled_task)
        assert unpickled_task.inputs[0].type.members
        self.assertEqual(unpickled_task.inputs[0].pos.line, 11)
        pickled_doc = pickle.dumps(doc)
        #self.assertLess(float(len(pickled_task))/len(pickled_doc), 0.6)

        doc = r"""
        version 1.0

        workflow wf {
            Person alyssa = object { name: "alyssa", "name": "ben" }
        }

        struct Person {
            String name
        }
        """
        with self.assertRaises(WDL.Error.MultipleDefinitions):
            doc = WDL.parse_document(doc)

        doc = r"""
        version 1.0

        workflow wf {
            Person alyssa = object { name: "alyssa" }
            Person ben = object { name: "ben", age: 42, bogus: 3.14}
            Person cy = object { name: "ben", age: false,}
            Int i = object {}
        }

        struct Person {
            String name
            Int age
        }
        """
        doc = WDL.parse_document(doc)
        with self.assertRaises(WDL.Error.MultipleValidationErrors) as ctx:
            doc.typecheck()
        self.assertEqual(len(ctx.exception.exceptions), 4)
        for i in range(4):
            self.assertTrue(isinstance(ctx.exception.exceptions[i], WDL.Error.StaticTypeMismatch))
        self.assertEqual(str(ctx.exception.exceptions[2]), "Expected Person instead of object(age : Boolean, name : String)")

        doc = r"""
        version 1.0

        workflow wf {
            Array[Person] ppl = [
                object { name: 'alyssa', friends: [2,4] },
                object { "name": "ben", 'friends': [8,16]},
                object { 'name': "cy", "friends": [32,64] }
            ]
        }

        struct Person {
            String name
            Array[Int] friends
        }
        """
        doc = WDL.parse_document(doc)
        doc.typecheck()

    def test_keywords(self):
        templ = r"""
        version 1.0

        struct {} {{
            String {}
        }}
        """
        WDL.parse_document(templ.format("foo","bar")).typecheck()
        for p in [
            ("task","bar"),
            ("foo","task"),
            ("struct","bar"),
            ("foo","struct"),
            ("Int","bar"),
            ("foo","Int")
        ]:
            try:
                WDL.parse_document(templ.format(*p))
                assert False
            except WDL.Error.SyntaxError as err:
                self.assertIsInstance(err.pos.line, int)
                self.assertIsInstance(err.pos.column, int)

class TestNoneLiteral(unittest.TestCase):
    def test_none_expr(self):
        doc = r"""
        version development
        struct Car {
            String make
            String model
        }
        workflow wf {
            input {
                Int? x = None
                Array[Float?] ax = [3, None]
                Array[Car?] ac = [None]
            }
            Array[Int?] a = [x, None]
            if (x == None) {
                Boolean y = true
            }
            output {
                Boolean b = defined(y)
            }
        }
        """
        doc = WDL.parse_document(doc)
        doc.typecheck()

        assert WDL.Value.Null() == WDL.Value.Null()
        assert str(WDL.Value.Null()) == "None"
        assert str(doc.workflow.inputs[0]) == "Int? x = None"

    def test_none_type_errors(self):
        with self.assertRaises(WDL.Error.StaticTypeMismatch):
            WDL.parse_document(r"""version development
            workflow w {
                Int x = None
            }
            """).typecheck()
        with self.assertRaises(WDL.Error.StaticTypeMismatch):
            WDL.parse_document(r"""version development
            workflow w {
                Array[Int] ax = [3, None]
            }
            """).typecheck()
        with self.assertRaises(WDL.Error.StaticTypeMismatch):
            doc = WDL.parse_document(r"""version development
            workflow w {
                Array[Float] ax = [3, None]
            }
            """)
            doc.typecheck()
            assert False, str(doc.workflow.body[0].expr.type)
        with self.assertRaises(WDL.Error.StaticTypeMismatch):
            WDL.parse_document(r"""version development
            struct Car {
                String make
                String model
            }
            workflow w {
                Array[Car] ax = [None]
            }
            """).typecheck()
        with self.assertRaises(WDL.Error.NoSuchMember):
            WDL.parse_document(r"""version development
            workflow w {
                Int x = None.left
            }
            """).typecheck()
        with self.assertRaises(WDL.Error.SyntaxError):
            WDL.parse_document(r"""version development
            struct Car {
                String make
                String model
            }
            workflow w {
                Car c
                String s = c.None
            }
            """).typecheck()
